# coding=utf-8
# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Portions Copyright (c) Microsoft Corporation

"""A library for tokenizing text."""

import re


# Pre-compile regexes that are use often
NON_ALPHANUM_PATTERN = r"[^a-z0-9]+"
NON_ALPHANUM_RE = re.compile(NON_ALPHANUM_PATTERN)
SPACES_PATTERN = r"\s+"
SPACES_RE = re.compile(SPACES_PATTERN)
VALID_TOKEN_PATTERN = r"^[a-z0-9]+$"
VALID_TOKEN_RE = re.compile(VALID_TOKEN_PATTERN)


def tokenize(text, stemmer):
    """Tokenize input text into a list of tokens.

    This approach aims to replicate the approach taken by Chin-Yew Lin in
    the original ROUGE implementation.

    Args:
      text: A text blob to tokenize.
      stemmer: An optional stemmer.

    Returns:
      A list of string tokens extracted from input text.
    """

    # Convert everything to lowercase.
    text = text.lower()
    # Replace any non-alpha-numeric characters with spaces.
    text = NON_ALPHANUM_RE.sub(" ", text)

    tokens = SPACES_RE.split(text)
    if stemmer:
        # Only stem words more than 3 characters long.
        tokens = [stemmer.stem(x) if len(x) > 3 else x for x in tokens]

    # One final check to drop any empty or invalid tokens.
    tokens = [x for x in tokens if VALID_TOKEN_RE.match(x)]

    return tokens



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import asyncio
import importlib.metadata
import math
import re
import time
import json
import html
from ast import literal_eval
from typing import Dict, List, Optional, Union, cast
from urllib.parse import urlparse
from string import Template
from azure.ai.evaluation._common.onedp._client import AIProjectClient
from azure.core.exceptions import HttpResponseError

import jwt

from azure.ai.evaluation._legacy._adapters._errors import MissingRequiredPackage
from azure.ai.evaluation._exceptions import ErrorBlame, ErrorCategory, ErrorTarget, EvaluationException
from azure.ai.evaluation._http_utils import AsyncHttpPipeline, get_async_http_client
from azure.ai.evaluation._model_configurations import AzureAIProject
from azure.ai.evaluation._common.utils import is_onedp_project
from azure.core.credentials import TokenCredential
from azure.core.exceptions import HttpResponseError
from azure.core.pipeline.policies import AsyncRetryPolicy

from .constants import (
    CommonConstants,
    EvaluationMetrics,
    RAIService,
    Tasks,
    _InternalEvaluationMetrics,
)
from .utils import get_harm_severity_level, retrieve_content_type

try:
    version = importlib.metadata.version("azure-ai-evaluation")
except importlib.metadata.PackageNotFoundError:
    version = "unknown"
USER_AGENT = "{}/{}".format("azure-ai-evaluation", version)

USER_TEXT_TEMPLATE_DICT: Dict[str, Template] = {
    "DEFAULT": Template("<Human>{$query}</><System>{$response}</>"),
}
ML_WORKSPACE  = "https://management.azure.com/.default"
COG_SRV_WORKSPACE = "https://ai.azure.com/.default"

INFERENCE_OF_SENSITIVE_ATTRIBUTES = "inference_sensitive_attributes"

def get_formatted_template(data: dict, annotation_task: str) -> str:
    """Given the task and input data, produce a formatted string that will serve as the main
    payload for the RAI service. Requires specific per-task logic.

    :param data: The data to incorporate into the payload.
    :type data: dict
    :param annotation_task: The annotation task to use. This determines the template to use.
    :type annotation_task: str
    :return: The formatted based on the data and task template.
    :rtype: str
    """
    # Template class doesn't play nice with json dumping/loading, just handle groundedness'
    # JSON format manually.
    # Template was: Template('{"question": "$query", "answer": "$response", "context": "$context"}'),
    if annotation_task == Tasks.GROUNDEDNESS:
        as_dict = {
            "question": data.get("query", ""),
            "answer": data.get("response", ""),
            "context": data.get("context", ""),
        }
        return json.dumps(as_dict)
    if annotation_task == Tasks.CODE_VULNERABILITY:
        as_dict = {
            "context": data.get("query", ""),
            "completion": data.get("response", "")
        }
        return json.dumps(as_dict)
    if annotation_task == Tasks.UNGROUNDED_ATTRIBUTES:
        as_dict = {
            "query": data.get("query", ""),
            "response": data.get("response", ""),
            "context": data.get("context", "")
        }
        return json.dumps(as_dict)
    as_dict = {
        "query": html.escape(data.get("query", "")),
        "response": html.escape(data.get("response", "")),
    }
    user_text = USER_TEXT_TEMPLATE_DICT.get(annotation_task, USER_TEXT_TEMPLATE_DICT["DEFAULT"]).substitute(**as_dict)
    return user_text.replace("'", '\\"')


def get_common_headers(token: str, evaluator_name: Optional[str] = None) -> Dict:
    """Get common headers for the HTTP request

    :param token: The Azure authentication token.
    :type token: str
    :param evaluator_name: The evaluator name. Default is None.
    :type evaluator_name: str
    :return: The common headers.
    :rtype: Dict
    """
    user_agent = f"{USER_AGENT} (type=evaluator; subtype={evaluator_name})" if evaluator_name else USER_AGENT
    return {
        "Authorization": f"Bearer {token}",
        "User-Agent": user_agent,
    }


def get_async_http_client_with_timeout() -> AsyncHttpPipeline:
    return get_async_http_client().with_policies(
        retry_policy=AsyncRetryPolicy(timeout=CommonConstants.DEFAULT_HTTP_TIMEOUT)
    )

async def ensure_service_availability_onedp(client: AIProjectClient, token: str, capability: Optional[str] = None) -> None:
    """Check if the Responsible AI service is available in the region and has the required capability, if relevant.

    :param client: The AI project client.
    :type client: AIProjectClient
    :param token: The Azure authentication token.
    :type token: str
    :param capability: The capability to check. Default is None.
    :type capability: str
    :raises Exception: If the service is not available in the region or the capability is not available.
    """
    headers = get_common_headers(token)
    capabilities = client.evaluations.check_annotation(headers=headers)
    
    if capability and capability not in capabilities:
        msg = f"The needed capability '{capability}' is not supported by the RAI service in this region."
        raise EvaluationException(
            message=msg,
            internal_message=msg,
            target=ErrorTarget.RAI_CLIENT,
            category=ErrorCategory.SERVICE_UNAVAILABLE,
            blame=ErrorBlame.USER_ERROR,
            tsg_link="https://aka.ms/azsdk/python/evaluation/safetyevaluator/troubleshoot",
        )
    
async def ensure_service_availability(rai_svc_url: str, token: str, capability: Optional[str] = None) -> None:
    """Check if the Responsible AI service is available in the region and has the required capability, if relevant.

    :param rai_svc_url: The Responsible AI service URL.
    :type rai_svc_url: str
    :param token: The Azure authentication token.
    :type token: str
    :param capability: The capability to check. Default is None.
    :type capability: str
    :raises Exception: If the service is not available in the region or the capability is not available.
    """
    headers = get_common_headers(token)
    svc_liveness_url = rai_svc_url + "/checkannotation"

    async with get_async_http_client() as client:
        response = await client.get(svc_liveness_url, headers=headers)

        if response.status_code != 200:
            msg = (
                f"RAI service is unavailable in this region, or you lack the necessary permissions "
                f"to access the AI project. Status Code: {response.status_code}"
            )
            raise EvaluationException(
                message=msg,
                internal_message=msg,
                target=ErrorTarget.RAI_CLIENT,
                category=ErrorCategory.SERVICE_UNAVAILABLE,
                blame=ErrorBlame.USER_ERROR,
                tsg_link="https://aka.ms/azsdk/python/evaluation/safetyevaluator/troubleshoot",
            )

        capabilities = response.json()
        if capability and capability not in capabilities:
            msg = f"The needed capability '{capability}' is not supported by the RAI service in this region."
            raise EvaluationException(
                message=msg,
                internal_message=msg,
                target=ErrorTarget.RAI_CLIENT,
                category=ErrorCategory.SERVICE_UNAVAILABLE,
                blame=ErrorBlame.USER_ERROR,
                tsg_link="https://aka.ms/azsdk/python/evaluation/safetyevaluator/troubleshoot",
            )


def generate_payload(normalized_user_text: str, metric: str, annotation_task: str) -> Dict:
    """Generate the payload for the annotation request

    :param normalized_user_text: The normalized user text to be entered as the "UserTextList" in the payload.
    :type normalized_user_text: str
    :param metric: The evaluation metric to use. This determines the task type, and whether a "MetricList" is needed
        in the payload.
    :type metric: str
    :param annotation_task: The annotation task to be passed to service
    :type annotation_task: str
    :return: The payload for the annotation request.
    :rtype: Dict
    """
    include_metric = True
    task = annotation_task
    if metric == EvaluationMetrics.PROTECTED_MATERIAL:
        include_metric = False
    elif metric == EvaluationMetrics.UNGROUNDED_ATTRIBUTES:
        include_metric = False
    elif metric == _InternalEvaluationMetrics.ECI:
        include_metric = False
    elif metric == EvaluationMetrics.XPIA:
        include_metric = False
    return (
        {
            "UserTextList": [normalized_user_text],
            "AnnotationTask": task,
            "MetricList": [metric],
        }
        if include_metric
        else {
            "UserTextList": [normalized_user_text],
            "AnnotationTask": task,
        }
    )


async def submit_request(
    data: dict, metric: str, rai_svc_url: str, token: str, annotation_task: str, evaluator_name: str
) -> str:
    """Submit request to Responsible AI service for evaluation and return operation ID

    :param data: The data to evaluate.
    :type data: dict
    :param metric: The evaluation metric to use.
    :type metric: str
    :param rai_svc_url: The Responsible AI service URL.
    :type rai_svc_url: str
    :param token: The Azure authentication token.
    :type token: str
    :param annotation_task: The annotation task to use.
    :type annotation_task: str
    :param evaluator_name: The evaluator name.
    :type evaluator_name: str
    :return: The operation ID.
    :rtype: str
    """
    normalized_user_text = get_formatted_template(data, annotation_task)
    payload = generate_payload(normalized_user_text, metric, annotation_task=annotation_task)

    url = rai_svc_url + "/submitannotation"
    headers = get_common_headers(token, evaluator_name)

    async with get_async_http_client_with_timeout() as client:
        http_response = await client.post(url, json=payload, headers=headers)

    if http_response.status_code != 202:
        print("Fail evaluating '%s' with error message: %s" % (payload["UserTextList"], http_response.text()))
        http_response.raise_for_status()
    result = http_response.json()
    operation_id = result["location"].split("/")[-1]
    return operation_id


async def submit_request_onedp(
    client: AIProjectClient, 
    data: dict, 
    metric: str, 
    token: str, 
    annotation_task: str, 
    evaluator_name: str
) -> str:
    """Submit request to Responsible AI service for evaluation and return operation ID

    :param client: The AI project client.
    :type client: AIProjectClient
    :param data: The data to evaluate.
    :type data: dict
    :param metric: The evaluation metric to use.
    :type metric: str
    :param token: The Azure authentication token.
    :type token: str
    :param annotation_task: The annotation task to use.
    :type annotation_task: str
    :param evaluator_name: The evaluator name.
    :type evaluator_name: str
    :return: The operation ID.
    :rtype: str
    """
    normalized_user_text = get_formatted_template(data, annotation_task)
    payload = generate_payload(normalized_user_text, metric, annotation_task=annotation_task)
    headers = get_common_headers(token, evaluator_name)
    response = client.evaluations.submit_annotation(payload, headers=headers)
    result = json.loads(response)
    operation_id = result["location"].split("/")[-1]
    return operation_id


async def fetch_result(operation_id: str, rai_svc_url: str, credential: TokenCredential, token: str) -> Dict:
    """Fetch the annotation result from Responsible AI service

    :param operation_id: The operation ID.
    :type operation_id: str
    :param rai_svc_url: The Responsible AI service URL.
    :type rai_svc_url: str
    :param credential: The Azure authentication credential.
    :type credential: ~azure.core.credentials.TokenCredential
    :param token: The Azure authentication token.
    :type token: str
    :return: The annotation result.
    :rtype: Dict
    """
    start = time.time()
    request_count = 0

    url = rai_svc_url + "/operations/" + operation_id
    while True:
        token = await fetch_or_reuse_token(credential, token)
        headers = get_common_headers(token)

        async with get_async_http_client_with_timeout() as client:
            response = await client.get(url, headers=headers)

        if response.status_code == 200:
            return response.json()

        request_count += 1
        time_elapsed = time.time() - start
        if time_elapsed > RAIService.TIMEOUT:
            raise TimeoutError(f"Fetching annotation result {request_count} times out after {time_elapsed:.2f} seconds")

        sleep_time = RAIService.SLEEP_TIME**request_count
        await asyncio.sleep(sleep_time)

async def fetch_result_onedp(client: AIProjectClient, operation_id: str, token: str) -> Dict:
    """Fetch the annotation result from Responsible AI service

    :param client: The AI project client.
    :type client: AIProjectClient
    :param operation_id: The operation ID.
    :type operation_id: str
    :param token: The Azure authentication token.
    :type token: str
    :return: The annotation result.
    :rtype: Dict
    """
    start = time.time()
    request_count = 0

    while True:
        headers = get_common_headers(token)
        try:
            return client.evaluations.operation_results(operation_id, headers=headers)
        except HttpResponseError:
            request_count += 1
            time_elapsed = time.time() - start
            if time_elapsed > RAIService.TIMEOUT:
                raise TimeoutError(f"Fetching annotation result {request_count} times out after {time_elapsed:.2f} seconds")

            sleep_time = RAIService.SLEEP_TIME**request_count
            await asyncio.sleep(sleep_time)

def parse_response(  # pylint: disable=too-many-branches,too-many-statements
    batch_response: List[Dict], metric_name: str, metric_display_name: Optional[str] = None
) -> Dict[str, Union[str, float]]:
    """Parse the annotation response from Responsible AI service for a content harm evaluation.

    :param batch_response: The annotation response from Responsible AI service.
    :type batch_response: List[Dict]
    :param metric_name: The evaluation metric to use.
    :type metric_name: str
    :param metric_display_name: The evaluation metric display name to use. If unset, use the metric_name.
    :type metric_display_name: Optional[str]
    :return: The parsed annotation result.
    :rtype: Dict[str, Union[str, float]]
    """
    if metric_display_name is None:
        metric_display_name = metric_name

    # non-numeric metrics
    if metric_name in {
        EvaluationMetrics.PROTECTED_MATERIAL,
        _InternalEvaluationMetrics.ECI,
        EvaluationMetrics.XPIA,
        EvaluationMetrics.CODE_VULNERABILITY,
        EvaluationMetrics.UNGROUNDED_ATTRIBUTES,
    }:
        result = {}
        if not batch_response or len(batch_response[0]) == 0:
            return {}
        if metric_name == EvaluationMetrics.UNGROUNDED_ATTRIBUTES and INFERENCE_OF_SENSITIVE_ATTRIBUTES in batch_response[0]:
            batch_response[0] = { 
                EvaluationMetrics.UNGROUNDED_ATTRIBUTES: batch_response[0][INFERENCE_OF_SENSITIVE_ATTRIBUTES] 
            } 
        if metric_name == EvaluationMetrics.PROTECTED_MATERIAL and metric_name not in batch_response[0]:
            pm_metric_names = {"artwork", "fictional_characters", "logos_and_brands"}
            for pm_metric_name in pm_metric_names:
                response = batch_response[0][pm_metric_name]
                response = response.replace("false", "False")
                response = response.replace("true", "True")
                parsed_response = literal_eval(response)
                result[pm_metric_name + "_label"] = parsed_response["label"] if "label" in parsed_response else math.nan
                result[pm_metric_name + "_reason"] = (
                    parsed_response["reasoning"] if "reasoning" in parsed_response else ""
                )
            return result
        if metric_name not in batch_response[0]:
            return {}
        response = batch_response[0][metric_name]
        response = response.replace("false", "False")
        response = response.replace("true", "True")
        parsed_response = literal_eval(response)
        # Use label instead of score since these are assumed to be boolean results.
        # Use math.nan as null value since it's ignored by aggregations rather than treated as 0.
        result[metric_display_name + "_label"] = parsed_response["label"] if "label" in parsed_response else math.nan
        result[metric_display_name + "_reason"] = parsed_response["reasoning"] if "reasoning" in parsed_response else ""

        if metric_name == EvaluationMetrics.XPIA:
            # Add "manipulated_content", "intrusion" and "information_gathering" to the result
            # if present else set them to math.nan
            result[metric_display_name + "_manipulated_content"] = (
                parsed_response["manipulated_content"] if "manipulated_content" in parsed_response else math.nan
            )
            result[metric_display_name + "_intrusion"] = (
                parsed_response["intrusion"] if "intrusion" in parsed_response else math.nan
            )
            result[metric_display_name + "_information_gathering"] = (
                parsed_response["information_gathering"] if "information_gathering" in parsed_response else math.nan
            )
        if metric_name == EvaluationMetrics.CODE_VULNERABILITY or metric_name == EvaluationMetrics.UNGROUNDED_ATTRIBUTES:
            # Add all attributes under the details.
            details = {}
            for key, value in parsed_response.items():
                if key not in {"label", "reasoning", "version"}:
                    details[key.replace("-", "_")] = value
            result[metric_display_name + "_details"] = details
        return result
    return _parse_content_harm_response(batch_response, metric_name, metric_display_name)


def _parse_content_harm_response(
    batch_response: List[Dict], metric_name: str, metric_display_name: Optional[str] = None
) -> Dict[str, Union[str, float]]:
    """Parse the annotation response from Responsible AI service for a content harm evaluation.

    :param batch_response: The annotation response from Responsible AI service.
    :type batch_response: List[Dict]
    :param metric_name: The evaluation metric to use.
    :type metric_name: str
    :param metric_display_name: The evaluation metric display name to use. If unset, use the metric_name.
    :type metric_display_name: Optional[str]
    :return: The parsed annotation result.
    :rtype: Dict[str, Union[str, float]]
    """
    # Fix the metric name if it's "hate_fairness"
    # Eventually we will remove this fix once the RAI service is updated
    key = metric_name if metric_display_name is None else metric_display_name
    if key == EvaluationMetrics.HATE_FAIRNESS:
        key = EvaluationMetrics.HATE_UNFAIRNESS

    result: Dict[str, Union[str, float]] = {
        (key.value if hasattr(key, "value") else key): math.nan,
        f"{(key.value if hasattr(key, 'value') else key)}_score": math.nan,
        f"{(key.value if hasattr(key, 'value') else key)}_reason": math.nan,
    }

    response = batch_response[0]
    if metric_name not in response:
        return result

    try:
        harm_response = literal_eval(response[metric_name])
    except Exception:  # pylint: disable=broad-exception-caught
        harm_response = response[metric_name]

    if harm_response != "" and isinstance(harm_response, dict):
        # check if "output" is one key in harm_response
        if "output" in harm_response:
            harm_response = harm_response["output"]

        # get content harm metric_value
        if "label" in harm_response:
            try:
                # Handle "n/a" or other non-numeric values
                if isinstance(harm_response["label"], str) and harm_response["label"].strip().lower() == "n/a":
                    metric_value = math.nan
                else:
                    metric_value = float(harm_response["label"])
            except (ValueError, TypeError):
                metric_value = math.nan
        elif "valid" in harm_response:
            metric_value = 0 if harm_response["valid"] else math.nan
        else:
            metric_value = math.nan

        # get reason
        if "reasoning" in harm_response:
            reason = harm_response["reasoning"]
        elif "reason" in harm_response:
            reason = harm_response["reason"]
        else:
            reason = ""
    elif harm_response != "" and isinstance(harm_response, str):
        metric_value_match = re.findall(r"(\b[0-7])\b", harm_response)
        if metric_value_match:
            metric_value = int(metric_value_match[0])
        else:
            metric_value = math.nan
        reason = harm_response
    elif harm_response != "" and isinstance(harm_response, (int, float)):
        if 0 < harm_response <= 7:
            metric_value = harm_response
        else:
            metric_value = math.nan
        reason = ""
    else:
        metric_value = math.nan
        reason = ""

    harm_score = metric_value
    # We've already handled the "n/a" case by converting to math.nan
    if not math.isnan(metric_value):
        # int(math.nan) causes a value error, and math.nan is already handled
        # by get_harm_severity_level
        harm_score = int(metric_value)
    result[key] = get_harm_severity_level(harm_score)
    result[key + "_score"] = harm_score
    result[key + "_reason"] = reason

    return result


async def _get_service_discovery_url(azure_ai_project: AzureAIProject, token: str) -> str:
    """Get the discovery service URL for the Azure AI project

    :param azure_ai_project: The Azure AI project details.
    :type azure_ai_project: ~azure.ai.evaluation.AzureAIProject
    :param token: The Azure authentication token.
    :type token: str
    :return: The discovery service URL.
    :rtype: str
    """
    headers = get_common_headers(token)

    async with get_async_http_client_with_timeout() as client:
        response = await client.get(
            f"https://management.azure.com/subscriptions/{azure_ai_project['subscription_id']}/"
            f"resourceGroups/{azure_ai_project['resource_group_name']}/"
            f"providers/Microsoft.MachineLearningServices/workspaces/{azure_ai_project['project_name']}?"
            f"api-version=2023-08-01-preview",
            headers=headers,
        )

    if response.status_code != 200:
        msg = (
            f"Failed to connect to your Azure AI project. Please check if the project scope is configured correctly, "
            f"and make sure you have the necessary access permissions. "
            f"Status code: {response.status_code}."
        )
        raise EvaluationException(
            message=msg,
            target=ErrorTarget.RAI_CLIENT,
            blame=ErrorBlame.USER_ERROR,
            category=ErrorCategory.PROJECT_ACCESS_ERROR,
            tsg_link="https://aka.ms/azsdk/python/evaluation/safetyevaluator/troubleshoot",
        )

    base_url = urlparse(response.json()["properties"]["discoveryUrl"])
    return f"{base_url.scheme}://{base_url.netloc}"


async def get_rai_svc_url(project_scope: AzureAIProject, token: str) -> str:
    """Get the Responsible AI service URL

    :param project_scope: The Azure AI project scope details.
    :type project_scope: Dict
    :param token: The Azure authentication token.
    :type token: str
    :return: The Responsible AI service URL.
    :rtype: str
    """
    discovery_url = await _get_service_discovery_url(azure_ai_project=project_scope, token=token)
    subscription_id = project_scope["subscription_id"]
    resource_group_name = project_scope["resource_group_name"]
    project_name = project_scope["project_name"]
    base_url = discovery_url.rstrip("/")
    rai_url = (
        f"{base_url}/raisvc/v1.0"
        f"/subscriptions/{subscription_id}"
        f"/resourceGroups/{resource_group_name}"
        f"/providers/Microsoft.MachineLearningServices/workspaces/{project_name}"
    )
    return rai_url


async def fetch_or_reuse_token(credential: TokenCredential, token: Optional[str] = None, workspace: Optional[str] = ML_WORKSPACE) -> str:
    """Get token. Fetch a new token if the current token is near expiry

       :param credential: The Azure authentication credential.
       :type credential:
    ~azure.core.credentials.TokenCredential
       :param token: The Azure authentication token. Defaults to None. If none, a new token will be fetched.
       :type token: str
       :return: The Azure authentication token.
    """
    if token:
        # Decode the token to get its expiration time
        try:
            decoded_token = jwt.decode(token, options={"verify_signature": False})
        except jwt.PyJWTError:
            pass
        else:
            exp_time = decoded_token["exp"]
            current_time = time.time()

            # Return current token if not near expiry
            if (exp_time - current_time) >= 300:
                return token

    return credential.get_token(workspace).token


async def evaluate_with_rai_service(
    data: dict,
    metric_name: str,
    project_scope: Union[str, AzureAIProject],
    credential: TokenCredential,
    annotation_task: str = Tasks.CONTENT_HARM,
    metric_display_name=None,
    evaluator_name=None,
) -> Dict[str, Union[str, float]]:
    """Evaluate the content safety of the response using Responsible AI service

    :param data: The data to evaluate.
    :type data: dict
    :param metric_name: The evaluation metric to use.
    :type metric_name: str
    :param project_scope: The Azure AI project, which can either be a string representing the project endpoint 
        or an instance of AzureAIProject. It contains subscription id, resource group, and project name. 
    :type project_scope: Union[str, AzureAIProject]
    :param credential: The Azure authentication credential.
    :type credential: ~azure.core.credentials.TokenCredential
    :param annotation_task: The annotation task to use.
    :type annotation_task: str
    :param metric_display_name: The display name of metric to use.
    :type metric_display_name: str
    :param evaluator_name: The evaluator name to use.
    :type evaluator_name: str
    :return: The parsed annotation result.
    :rtype: Dict[str, Union[str, float]]
    """

    if is_onedp_project(project_scope):
        client = AIProjectClient(endpoint=project_scope, credential=credential)
        token = await fetch_or_reuse_token(credential=credential, workspace=COG_SRV_WORKSPACE)
        await ensure_service_availability_onedp(client, token, annotation_task)
        operation_id = await submit_request_onedp(client, data, metric_name, token, annotation_task, evaluator_name)
        annotation_response = cast(List[Dict], await fetch_result_onedp(client, operation_id, token))
        result = parse_response(annotation_response, metric_name, metric_display_name)
        return result
    else:
        # Get RAI service URL from discovery service and check service availability
        token = await fetch_or_reuse_token(credential)
        rai_svc_url = await get_rai_svc_url(project_scope, token)
        await ensure_service_availability(rai_svc_url, token, annotation_task)

        # Submit annotation request and fetch result
        operation_id = await submit_request(data, metric_name, rai_svc_url, token, annotation_task, evaluator_name)
        annotation_response = cast(List[Dict], await fetch_result(operation_id, rai_svc_url, credential, token))
        result = parse_response(annotation_response, metric_name, metric_display_name)

        return result

def generate_payload_multimodal(content_type: str, messages, metric: str) -> Dict:
    """Generate the payload for the annotation request
    :param content_type: The type of the content representing multimodal or images.
    :type content_type: str
    :param messages: The normalized list of messages to be entered as the "Contents" in the payload.
    :type messages: str
    :param metric: The evaluation metric to use. This determines the task type, and whether a "MetricList" is needed
        in the payload.
    :type metric: str
    :return: The payload for the annotation request.
    :rtype: Dict
    """
    include_metric = True
    task = Tasks.CONTENT_HARM
    if metric == EvaluationMetrics.PROTECTED_MATERIAL:
        task = Tasks.PROTECTED_MATERIAL
        include_metric = False

    if include_metric:
        return {
            "ContentType": content_type,
            "Contents": [{"messages": messages}],
            "AnnotationTask": task,
            "MetricList": [metric],
        }
    return {
        "ContentType": content_type,
        "Contents": [{"messages": messages}],
        "AnnotationTask": task,
    }

async def submit_multimodal_request(messages, metric: str, rai_svc_url: str, token: str) -> str:
    """Submit request to Responsible AI service for evaluation and return operation ID
    :param messages: The normalized list of messages to be entered as the "Contents" in the payload.
    :type messages: str
    :param metric: The evaluation metric to use.
    :type metric: str
    :param rai_svc_url: The Responsible AI service URL.
    :type rai_svc_url: str
    :param token: The Azure authentication token.
    :type token: str
    :return: The operation ID.
    :rtype: str
    """
    ## handle json payload and payload from inference sdk strongly type messages
    if len(messages) > 0 and not isinstance(messages[0], dict):
        try:
            from azure.ai.inference.models import ChatRequestMessage
        except ImportError as ex:
            error_message = (
                "Please install 'azure-ai-inference' package to use SystemMessage, UserMessage, AssistantMessage"
            )
            raise MissingRequiredPackage(message=error_message) from ex
        if len(messages) > 0 and isinstance(messages[0], ChatRequestMessage):
            messages = [message.as_dict() for message in messages]

    filtered_messages = [message for message in messages if message["role"] != "system"]
    assistant_messages = [message for message in messages if message["role"] == "assistant"]
    content_type = retrieve_content_type(assistant_messages, metric)
    payload = generate_payload_multimodal(content_type, filtered_messages, metric)

    ## calling rai service for annotation
    url = rai_svc_url + "/submitannotation"
    headers = get_common_headers(token)
    async with get_async_http_client() as client:
        response = await client.post(  # pylint: disable=too-many-function-args,unexpected-keyword-arg
            url, json=payload, headers=headers
        )
    if response.status_code != 202:
        raise HttpResponseError(
            message=f"Received unexpected HTTP status: {response.status_code} {response.text()}", response=response
        )
    result = response.json()
    operation_id = result["location"].split("/")[-1]
    return operation_id

async def submit_multimodal_request_onedp(client: AIProjectClient, messages, metric: str, token: str) -> str:
    
    #  handle inference sdk strongly type messages
    if len(messages) > 0 and not isinstance(messages[0], dict):
        try:
            from azure.ai.inference.models import ChatRequestMessage
        except ImportError as ex:
            error_message = (
                "Please install 'azure-ai-inference' package to use SystemMessage, UserMessage, AssistantMessage"
            )
            raise MissingRequiredPackage(message=error_message) from ex
        if len(messages) > 0 and isinstance(messages[0], ChatRequestMessage):
            messages = [message.as_dict() for message in messages]

    ## fetch system and assistant messages from the list of messages
    filtered_messages = [message for message in messages if message["role"] != "system"]
    assistant_messages = [message for message in messages if message["role"] == "assistant"]
    
    ## prepare for request
    content_type = retrieve_content_type(assistant_messages, metric)
    payload = generate_payload_multimodal(content_type, filtered_messages, metric)
    headers = get_common_headers(token)
    
    response = client.evaluations.submit_annotation(payload, headers=headers)
    
    result = json.loads(response)
    operation_id = result["location"].split("/")[-1]
    return operation_id

async def evaluate_with_rai_service_multimodal(
    messages, metric_name: str, project_scope: Union[str, AzureAIProject], credential: TokenCredential
):
    """ "Evaluate the content safety of the response using Responsible AI service
       :param messages: The normalized list of messages.
       :type messages: str
       :param metric_name: The evaluation metric to use.
       :type metric_name: str
       :param project_scope: The Azure AI project, which can either be a string representing the project endpoint 
            or an instance of AzureAIProject. It contains subscription id, resource group, and project name. 
       :type project_scope: Union[str, AzureAIProject]
       :param credential: The Azure authentication credential.
       :type credential: ~azure.core.credentials.TokenCredential
       :return: The parsed annotation result.
       :rtype: List[List[Dict]]
    """

    if is_onedp_project(project_scope):
        client = AIProjectClient(endpoint=project_scope, credential=credential)
        token = await fetch_or_reuse_token(credential=credential, workspace=COG_SRV_WORKSPACE)
        await ensure_service_availability_onedp(client, token, Tasks.CONTENT_HARM)
        operation_id = await submit_multimodal_request_onedp(client, messages, metric_name, token)
        annotation_response = cast(List[Dict], await fetch_result_onedp(client, operation_id, token))
        result = parse_response(annotation_response, metric_name)
        return result
    else:
        token = await fetch_or_reuse_token(credential)
        rai_svc_url = await get_rai_svc_url(project_scope, token)
        await ensure_service_availability(rai_svc_url, token, Tasks.CONTENT_HARM)
        # Submit annotation request and fetch result
        operation_id = await submit_multimodal_request(messages, metric_name, rai_svc_url, token)
        annotation_response = cast(List[Dict], await fetch_result(operation_id, rai_svc_url, credential, token))
        result = parse_response(annotation_response, metric_name)
        return result


from unittest.mock import MagicMock

import pytest
from azure.ai.evaluation import ResponseCompletenessEvaluator
from azure.ai.evaluation._exceptions import EvaluationException


async def completeness_response1_async_mock():
    return """<S0>Let's think step by step: The ground truth states "The capital of Japan is Tokyo," which provides 
    both the subject (capital of Japan) and the specific answer (Tokyo). The response, "The capital of Japan,
    " only partially addresses the subject but does not provide the specific answer (Tokyo). This means it misses the 
    core claim established in the ground truth.</S0> <S1>The response is fully incomplete as it does not provide the 
    necessary and relevant information, specifically the name of the capital, Tokyo.</S1> <S2>1</S2>"""


async def completeness_response2_async_mock():
    return """<S0>Let's think step by step: The response states that the capital of Japan is Tokyo. The ground truth 
    also states that the capital of Japan is Tokyo. Both the response and the ground truth are identical, containing 
    all the necessary and relevant information. There is no missing or incorrect information in the response.</S0> 
    <S1>The response perfectly matches the ground truth, containing all the necessary and relevant information 
    without any omissions or errors.</S1> <S2>5</S2>"""


@pytest.mark.usefixtures("mock_model_config")
@pytest.mark.unittest
class TestResponseCompletenessEvaluator:
    def test_initialization(self, mock_model_config):
        response_completeness_evaluator = ResponseCompletenessEvaluator(model_config=mock_model_config)
        # Test initialization of ResponseCompletenessEvaluator
        assert response_completeness_evaluator.threshold == ResponseCompletenessEvaluator._DEFAULT_COMPLETENESS_THRESHOLD
        assert response_completeness_evaluator._result_key == ResponseCompletenessEvaluator._RESULT_KEY
        assert response_completeness_evaluator._is_reasoning_model is False

    def test_initialization2(self, mock_model_config):
        response_completeness_evaluator = ResponseCompletenessEvaluator(model_config=mock_model_config,
                                                                        is_reasoning_model=True)
        # Test initialization of ResponseCompletenessEvaluator
        assert response_completeness_evaluator.threshold == ResponseCompletenessEvaluator._DEFAULT_COMPLETENESS_THRESHOLD
        assert response_completeness_evaluator._result_key == ResponseCompletenessEvaluator._RESULT_KEY
        assert response_completeness_evaluator._is_reasoning_model is True

    def test_evaluate_completeness_valid1(self, mock_model_config):
        response_completeness_evaluator = ResponseCompletenessEvaluator(model_config=mock_model_config)
        response_completeness_evaluator._flow = MagicMock(return_value=completeness_response1_async_mock())

        # Test evaluation with valid ground truth and response
        ground_truth = "The capital of Japan is Tokyo."
        response = "The capital of Japan"
        result = response_completeness_evaluator(ground_truth=ground_truth, response=response)

        key = ResponseCompletenessEvaluator._RESULT_KEY
        assert result is not None
        assert (key in result and f"{key}_result" in result and f"{key}_threshold" in result
                and f"{key}_reason" in result)
        assert result[key] == 1
        assert result[f"{key}_result"] == "fail"
        assert result[f"{key}_threshold"] == ResponseCompletenessEvaluator._DEFAULT_COMPLETENESS_THRESHOLD
        assert "The response is fully incomplete " in result[f"{key}_reason"]

    def test_evaluate_completeness_valid2(self, mock_model_config):
        response_completeness_evaluator = ResponseCompletenessEvaluator(model_config=mock_model_config)
        response_completeness_evaluator._flow = MagicMock(return_value=completeness_response2_async_mock())

        # Test evaluation with valid ground truth and response
        ground_truth = "The capital of Japan is Tokyo."
        response = "The capital of Japan is Tokyo."
        result = response_completeness_evaluator(ground_truth=ground_truth, response=response)

        key = ResponseCompletenessEvaluator._RESULT_KEY
        assert result is not None

        assert (key in result and f"{key}_result" in result and f"{key}_threshold" in result and
                f"{key}_reason" in result)
        assert result[key] == 5
        assert result[f"{key}_result"] == "pass"
        assert result[f"{key}_threshold"] == ResponseCompletenessEvaluator._DEFAULT_COMPLETENESS_THRESHOLD
        assert "The response perfectly matches " in result[f"{key}_reason"]

    def test_evaluate_completeness_valid3(self, mock_model_config):
        response_completeness_evaluator = ResponseCompletenessEvaluator(model_config=mock_model_config,
                                                                        is_reasoning_model=True)
        response_completeness_evaluator._flow = MagicMock(return_value=completeness_response2_async_mock())

        # Test evaluation with valid ground truth and response
        ground_truth = "The capital of Japan is Tokyo."
        response = "The capital of Japan is Tokyo."
        result = response_completeness_evaluator(ground_truth=ground_truth, response=response)

        key = ResponseCompletenessEvaluator._RESULT_KEY
        assert result is not None

        assert (key in result and f"{key}_result" in result and f"{key}_threshold" in result and
                f"{key}_reason" in result)
        assert result[key] == 5
        assert result[f"{key}_result"] == "pass"
        assert result[f"{key}_threshold"] == ResponseCompletenessEvaluator._DEFAULT_COMPLETENESS_THRESHOLD
        assert "The response perfectly matches " in result[f"{key}_reason"]

    def test_evaluate_completeness_missing_ground_truth(self, mock_model_config):
        response_completeness_evaluator = ResponseCompletenessEvaluator(model_config=mock_model_config)
        response_completeness_evaluator._flow = MagicMock(return_value=completeness_response1_async_mock())

        # Test evaluation with missing ground truth
        response = "The capital of China is Beijing."
        with pytest.raises(EvaluationException) as exc_info:
            response_completeness_evaluator(response=response)

        assert ("ResponseCompletenessEvaluator: Either 'conversation' or individual inputs must be provided."
                in exc_info.value.args[0])

    def test_evaluate_completeness_missing_response(self, mock_model_config):
        response_completeness_evaluator = ResponseCompletenessEvaluator(model_config=mock_model_config)
        response_completeness_evaluator._flow = MagicMock(return_value=completeness_response1_async_mock())

        # Test evaluation with missing ground truth
        ground_truth = "The capital of China is Beijing."
        with pytest.raises(EvaluationException) as exc_info:
            response_completeness_evaluator(ground_truth=ground_truth)

        assert ("ResponseCompletenessEvaluator: Either 'conversation' or individual inputs must be provided."
                in exc_info.value.args[0])



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Any


class AttrDict(dict):
    """A dictionary that allows attribute access to its keys."""

    def __getattr__(self, key: str) -> Any:
        return self[key]

    def __setattr__(self, key: str, value: Any) -> None:
        self[key] = value



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from os import PathLike
from typing import IO, Any, Dict, Optional, Union, cast

from ruamel.yaml import YAML, YAMLError  # cspell:ignore ruamel

from azure.ai.evaluation._constants import DefaultOpenEncoding
from azure.ai.evaluation._legacy.prompty._exceptions import MissingRequiredInputError


def load_yaml(source: Optional[Union[str, PathLike, IO]]) -> Dict:
    # null check - just return an empty dict.
    # Certain CLI commands rely on this behavior to produce a resource
    # via CLI, which is then populated through CLArgs.
    """Load a local YAML file or a readable stream object.

    .. note::

        1. For a local file yaml

        .. code-block:: python

            yaml_path = "path/to/yaml"
            content = load_yaml(yaml_path)

        2. For a readable stream object

        .. code-block:: python

            with open("path/to/yaml", "r", encoding="utf-8") as f:
                content = load_yaml(f)


    :param source: The relative or absolute path to the local file, or a readable stream object.
    :type source: str
    :return: A dictionary representation of the local file's contents.
    :rtype: Dict
    """

    if source is None:
        return {}

    # pylint: disable=redefined-builtin
    input: Optional[IO] = None
    must_open_file = False
    try:  # check source type by duck-typing it as an IOBase
        readable = cast(IO, source).readable()
        if not readable:  # source is misformatted stream or file
            msg = "File Permissions Error: The already-open \n\n inputted file is not readable."
            raise PermissionError(msg)
        # source is an already-open stream or file, we can read() from it directly.
        input = cast(IO, source)
    except AttributeError:
        # source has no writable() function, assume it's a string or file path.
        must_open_file = True

    if must_open_file:  # If supplied a file path, open it.
        try:
            input = open(  # pylint: disable=consider-using-with
                cast(Union[PathLike, str], source), "r", encoding=DefaultOpenEncoding.READ
            )
        except OSError:  # FileNotFoundError introduced in Python 3
            e = FileNotFoundError(f"No such file or directory: {source}")
            raise MissingRequiredInputError(str(e), privacy_info=[str(source)]) from e
    # input should now be a readable file or stream. Parse it.
    try:
        yaml = YAML()
        yaml.preserve_quotes = True
        cfg = yaml.load(input)
    except YAMLError as e:
        msg = f"Error while parsing yaml file: {source} \n\n {str(e)}"
        raise YAMLError(msg) from e
    finally:
        if input and must_open_file:
            input.close()

    return cfg or {}


def load_yaml_string(yaml_string: str) -> Dict[str, Any]:
    """Load a yaml string.

    .. code-block:: python

        yaml_string = "some yaml string"
        object = load_yaml_string(yaml_string)


    :param yaml_string: A yaml string.
    :type yaml_string: str
    :return: A dictionary representation of the yaml string.
    :rtype: Dict
    """
    yaml = YAML()
    yaml.preserve_quotes = True
    return yaml.load(yaml_string)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from azure.ai.evaluation._model_configurations import AzureOpenAIModelConfiguration, OpenAIModelConfiguration

from azure.ai.evaluation._constants import DEFAULT_AOAI_API_VERSION
from azure.ai.evaluation._exceptions import ErrorBlame, ErrorCategory, ErrorTarget, EvaluationException
from typing import Any, Dict, Union
from azure.ai.evaluation._common._experimental import experimental


@experimental
class AzureOpenAIGrader():
    """
    Base class for Azure OpenAI grader wrappers, recommended only for use by experienced OpenAI API users.
    Combines a model configuration and any grader configuration
    into a singular object that can be used in evaluations.

    Supplying an AzureOpenAIGrader to the `evaluate` method will cause an asynchronous request to evaluate
    the grader via the OpenAI API. The results of the evaluation will then be merged into the standard
    evaluation results.

    :param model_config: The model configuration to use for the grader.
    :type model_config: Union[
        ~azure.ai.evaluation.AzureOpenAIModelConfiguration,
        ~azure.ai.evaluation.OpenAIModelConfiguration
    ]
    :param grader_config: The grader configuration to use for the grader. This is expected
        to be formatted as a dictionary that matches the specifications of the sub-types of
        the TestingCriterion alias specified in (OpenAI's SDK)[https://github.com/openai/openai-python/blob/ed53107e10e6c86754866b48f8bd862659134ca8/src/openai/types/eval_create_params.py#L151].
    :type grader_config: Dict[str, Any]
    :param kwargs: Additional keyword arguments to pass to the grader.
    :type kwargs: Any


    """

    id = "aoai://general"

    def __init__(self, *, model_config : Union[AzureOpenAIModelConfiguration, OpenAIModelConfiguration], grader_config: Dict[str, Any], **kwargs: Any):
        self._model_config = model_config
        self._grader_config = grader_config

        if kwargs.get("validate", True):
            self._validate_model_config()
            self._validate_grader_config()



    def _validate_model_config(self) -> None:
        """Validate the model configuration that this grader wrapper is using."""
        if "api_key" not in self._model_config or not self._model_config.get("api_key"):
            msg = f"{type(self).__name__}: Requires an api_key in the supplied model_config."
            raise EvaluationException(
                message=msg,
                blame=ErrorBlame.USER_ERROR,
                category=ErrorCategory.INVALID_VALUE,
                target=ErrorTarget.AOAI_GRADER,
            )
    
    def _validate_grader_config(self) -> None:
        """Validate the grader configuration that this grader wrapper is using."""

        return

    def get_client(self) -> Any:
        """Construct an appropriate OpenAI client using this grader's model configuration.
        Returns a slightly different client depending on whether or not this grader's model
        configuration is for Azure OpenAI or OpenAI.

        :return: The OpenAI client.
        :rtype: [~openai.OpenAI, ~openai.AzureOpenAI]
        """
        if "azure_endpoint" in self._model_config:
           from openai import AzureOpenAI
           # TODO set default values?
           return AzureOpenAI(
                azure_endpoint=self._model_config["azure_endpoint"],
                api_key=self._model_config.get("api_key", None), # Default-style access to appease linters.
                api_version=DEFAULT_AOAI_API_VERSION, # Force a known working version
                azure_deployment=self._model_config.get("azure_deployment", ""),
            )
        from openai import OpenAI
        # TODO add default values for base_url and organization?
        return OpenAI(
            api_key=self._model_config["api_key"],
            base_url=self._model_config.get("base_url", ""),
            organization=self._model_config.get("organization", ""),
        )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from os import PathLike
from typing import Any, Callable, Dict, Optional, Union
from typing_extensions import TypeAlias

import pandas as pd

from ._errors import MissingRequiredPackage
from ._configuration import Configuration
from .entities import Run


try:
    from promptflow.client import PFClient as _PFClient
except ImportError:

    class _PFClient:
        def __init__(self, **kwargs):
            self._config = Configuration(override_config=kwargs.pop("config", None))

        def run(
            self,
            flow: Union[str, PathLike, Callable],
            *,
            data: Union[str, PathLike],
            run: Optional[Union[str, Run]] = None,
            column_mapping: Optional[dict] = None,
            variant: Optional[str] = None,
            connections: Optional[dict] = None,
            environment_variables: Optional[dict] = None,
            name: Optional[str] = None,
            display_name: Optional[str] = None,
            tags: Optional[Dict[str, str]] = None,
            resume_from: Optional[Union[str, Run]] = None,
            code: Optional[Union[str, PathLike]] = None,
            init: Optional[dict] = None,
            **kwargs,
        ) -> Run:
            raise MissingRequiredPackage("Please install 'promptflow' package to use PFClient")

        def get_details(self, run: Union[str, Run], max_results: int = 100, all_results: bool = False) -> pd.DataFrame:
            return pd.DataFrame()

        def get_metrics(self, run: Union[str, Run]) -> Dict[str, Any]:
            return {}


PFClient: TypeAlias = _PFClient



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing import List, Union, Dict
from typing_extensions import overload, override

from azure.ai.evaluation._common._experimental import experimental
from azure.ai.evaluation._common.constants import EvaluationMetrics
from azure.ai.evaluation._evaluators._common import RaiServiceEvaluatorBase
from azure.ai.evaluation._model_configurations import Conversation


@experimental
class GroundednessProEvaluator(RaiServiceEvaluatorBase[Union[str, bool]]):
    """
    Evaluates service-based groundedness score for a given response, context, and query or a multi-turn conversation,
    including reasoning.

    The groundedness measure calls Azure AI Evaluation service to assess how well the AI-generated answer is grounded
    in the source context. Even if the responses from LLM are factually correct, they'll be considered ungrounded if
    they can't be verified against the provided sources (such as your input source or your database).

    Service-based groundedness scores are boolean values, where True indicates that the response is grounded.

    :param credential: The credential for connecting to Azure AI project. Required
    :type credential: ~azure.core.credentials.TokenCredential
    :param azure_ai_project: The scope of the Azure AI project.
        It contains subscription id, resource group, and project name.
    :type azure_ai_project: ~azure.ai.evaluation.AzureAIProject
    :param threshold: The threshold for the groundedness pro evaluator. Default is 5.
    :type threshold: int
    :param kwargs: Additional arguments to pass to the evaluator.
    :type kwargs: Any

    .. admonition:: Example:

        .. literalinclude:: ../samples/evaluation_samples_evaluate.py
            :start-after: [START groundedness_pro_evaluator]
            :end-before: [END groundedness_pro_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call a GroundednessProEvaluator with a query, response, and context.

    .. admonition:: Example using Azure AI Project URL:
        
        .. literalinclude:: ../samples/evaluation_samples_evaluate_fdp.py
            :start-after: [START groundedness_pro_evaluator]
            :end-before: [END groundedness_pro_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call GroundednessProEvaluator using Azure AI Project URL in the following format 
                https://{resource_name}.services.ai.azure.com/api/projects/{project_name}

    .. admonition:: Example with threshold:

        .. literalinclude:: ../samples/evaluation_samples_threshold.py
            :start-after: [START threshold_groundedness_pro_evaluator]
            :end-before: [END threshold_groundedness_pro_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize with a specified threshold and call GroundednessProEvaluator with a query, response, and context.

    .. note::

        If this evaluator is supplied to the `evaluate` function, the aggregated metric
        for the groundedness pro label will be "groundedness_pro_passing_rate".
    """

    id = "azureml://registries/azureml/models/Groundedness-Pro-Evaluator/versions/1"
    """Evaluator identifier, experimental and to be used only with evaluation in cloud."""

    @override
    def __init__(
        self,
        credential,
        azure_ai_project,
        *,
        threshold: int = 5,
        **kwargs,
    ):
        self.threshold = threshold
        self._higher_is_better = True
        self._output_prefix = "groundedness_pro"
        super().__init__(
            eval_metric=EvaluationMetrics.GROUNDEDNESS,
            azure_ai_project=azure_ai_project,
            credential=credential,
            threshold=self.threshold,
            **kwargs,
        )

    @overload
    def __call__(
        self,
        *,
        response: str,
        context: str,
        query: str,
    ) -> Dict[str, Union[str, bool]]:
        """Evaluate groundedness for a given query/response/context

        :keyword response: The response to be evaluated.
        :paramtype response: str
        :keyword context: The context to be evaluated.
        :paramtype context: str
        :keyword query: The query to be evaluated.
        :paramtype query: Optional[str]
        :return: The relevance score.
        :rtype: Dict[str, Union[str, bool]]
        """

    @overload
    def __call__(
        self,
        *,
        conversation: Conversation,
    ) -> Dict[str, Union[float, Dict[str, List[Union[str, bool]]]]]:
        """Evaluate groundedness for a conversation for a multi-turn evaluation. If the conversation has
        more than one turn, the evaluator will aggregate the results of each turn, with the per-turn results
        available in the output under the "evaluation_per_turn" key.

        :keyword conversation: The conversation to evaluate. Expected to contain a list of conversation turns under the
            key "messages", and potentially a global context under the key "context". Conversation turns are expected
            to be dictionaries with keys "content", "role", and possibly "context".
        :paramtype conversation: Optional[~azure.ai.evaluation.Conversation]
        :return: The relevance score.
        :rtype: Dict[str, Union[float, Dict[str, List[Union[str, bool]]]]]
        """

    @override
    def __call__(  # pylint: disable=docstring-missing-param
        self,
        *args,
        **kwargs,
    ):
        """Evaluate groundedness. Accepts either a query, response and context for a single-turn evaluation, or a
        or a conversation for a multi-turn evaluation. If the conversation has more than one turn,
        the evaluator will aggregate the results of each turn, with the per-turn results available
        in the output under the "evaluation_per_turn" key.

        :keyword query: The query to be evaluated.
        :paramtype query: Optional[str]
        :keyword response: The response to be evaluated.
        :paramtype response: Optional[str]
        :keyword context: The context to be evaluated.
        :paramtype context: Optional[str]
        :keyword conversation: The conversation to evaluate. Expected to contain a list of conversation turns under the
            key "messages", and potentially a global context under the key "context". Conversation turns are expected
            to be dictionaries with keys "content", "role", and possibly "context".
        :paramtype conversation: Optional[~azure.ai.evaluation.Conversation]
        :return: The relevance score.
        :rtype: Union[Dict[str, Union[str, bool]], Dict[str, Union[float, Dict[str, List[Union[str, bool]]]]]]
        """
        return super().__call__(*args, **kwargs)

    @override
    async def _do_eval(self, eval_input: Dict):
        """This evaluator has some unique post-processing that requires data that
        the rai_service script is not currently built to handle. So we post-post-process
        the result here to message it into the right form.

        :param eval_input: The input to the evaluation function.
        :type eval_input: Dict
        :return: The evaluation result.
        :rtype: Dict
        """
        result = await super()._do_eval(eval_input)
        real_result = {}
        real_result[self._output_prefix + "_reason"] = result[EvaluationMetrics.GROUNDEDNESS + "_reason"]
        real_result[self._output_prefix + "_label"] = (
            result[EvaluationMetrics.GROUNDEDNESS + "_score"] >= self.threshold
        )
        if self._higher_is_better:
            real_result[self._output_prefix + "_score"] = max(result[EvaluationMetrics.GROUNDEDNESS + "_score"], 0)
        else:
            real_result[self._output_prefix + "_score"] = min(result[EvaluationMetrics.GROUNDEDNESS + "_score"], 1)

        return real_result



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import time
from typing import Dict
from typing_extensions import overload, override


from azure.ai.evaluation._evaluators._common import EvaluatorBase


class SlowEvaluator(EvaluatorBase[str]):
    """Test evaluator that just returns the input after a slight delay. Used for testing performance."""

    def __call__(
        self,
        *,
        query: str,
    ) -> Dict[str, str]:
        """Evaluate a collection of content safety metrics for the given query/response pair

        :keyword query: The query to be evaluated.
        :paramtype query: str
        :keyword response: The response to be evaluated.
        :paramtype response: str
        :return: The content safety scores.
        :rtype: Dict[str, Union[str, float]]
        """
        return super().__call__(query=query)

    @override
    async def _do_eval(self, eval_input: Dict) -> Dict[str, str]:
        time.sleep(0.5)
        return {"result": "done"}



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing import Any, Dict, Union
from typing_extensions import Literal

from azure.ai.evaluation._model_configurations import AzureOpenAIModelConfiguration, OpenAIModelConfiguration
from openai.types.graders import StringCheckGrader
from azure.ai.evaluation._common._experimental import experimental

from .aoai_grader import AzureOpenAIGrader

@experimental
class AzureOpenAIStringCheckGrader(AzureOpenAIGrader):
    """
    Wrapper class for OpenAI's string check graders.

    Supplying a StringCheckGrader to the `evaluate` method will cause an asynchronous request to evaluate
    the grader via the OpenAI API. The results of the evaluation will then be merged into the standard
    evaluation results.

    :param model_config: The model configuration to use for the grader.
    :type model_config: Union[
        ~azure.ai.evaluation.AzureOpenAIModelConfiguration,
        ~azure.ai.evaluation.OpenAIModelConfiguration
    ]
    :param input: The input text. This may include template strings.
    :type input: str
    :param name: The name of the grader.
    :type name: str
    :param operation: The string check operation to perform. One of `eq`, `ne`, `like`, or `ilike`.
    :type operation: Literal["eq", "ne", "like", "ilike"]
    :param reference: The reference text. This may include template strings.
    :type reference: str
    :param kwargs: Additional keyword arguments to pass to the grader.
    :type kwargs: Any


    """

    id = "aoai://string_check"

    def __init__(
        self,
        *,
        model_config : Union[AzureOpenAIModelConfiguration, OpenAIModelConfiguration],
        input: str,
        name: str,
        operation: Literal[
            "eq",
            "ne",
            "like",
            "ilike",
        ],
        reference: str,
        **kwargs: Any
    ):
        grader = StringCheckGrader(
            input=input,
            name=name,
            operation=operation,
            reference=reference,
            type="string_check",
        )
        super().__init__(model_config=model_config, grader_config=grader, **kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# cspell:ignore apng, retriable

import copy
import os
import re
import json
import base64
from dataclasses import dataclass, is_dataclass, fields
from logging import Logger
from pathlib import Path
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    Final,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from jinja2 import Template
from openai import AsyncStream
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai import APIConnectionError, APIStatusError, APITimeoutError, OpenAIError

from azure.ai.evaluation._constants import DefaultOpenEncoding
from azure.ai.evaluation._legacy.prompty._exceptions import (
    InvalidInputError,
    JinjaTemplateError,
    PromptyException,
)

from azure.ai.evaluation._legacy.prompty._yaml_utils import load_yaml


# region: Resolving references


@dataclass
class PromptyModelConfiguration:
    """
    A dataclass that represents a model config of prompty.

    :param api: Type of the LLM request, default value is chat.
    :type api: str
    :param configuration: Prompty model connection configuration
    :type configuration: dict
    :param parameters: Params of the LLM request.
    :type parameters: dict
    :param response: Return the complete response or the first choice, default value is first.
    :type response: str
    """

    configuration: dict
    parameters: Dict[str, Any]
    response: str = "first"
    model: Optional[str] = None
    # _overflow: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        if not isinstance(self.configuration, dict):
            raise PromptyException("The configuration of the model must be a dictionary.")

        if not self.model:
            self.model = self.configuration.get("azure_deployment", None) or self.configuration.get("model", None)


T = TypeVar("T")


def dataclass_from_dict(cls: Type[T], data: Dict[str, Any]) -> T:
    """Helper function to make creating dataclass instances from dictionaries easier.
    Unlike using cls(**data), this function will ignore any keys in the dictionary that
    are not fields in the dataclass. If the dataclass optionally contains an _overflow
    field, any extra key/value paris will be placed in that field.

    This does no type checking and inspects only the key names.

    :param Type[T] cls: The dataclass type to create.
    :param Dict[str, Any] data: The dictionary to create the dataclass instance from.
    :return: The dataclass instance.
    :rtype: T
    """
    if not is_dataclass(cls):
        raise ValueError("This function only works with @dataclass Types")

    fields_set: Set[str] = {f.name for f in fields(cls)}

    params: Dict[str, Any] = {}
    overflow: Dict[str, Any] = {}

    for key, value in data.items():
        if key in fields_set:
            params[key] = value
        else:
            overflow[key] = value

    if "_overflow" in fields_set:
        params["_overflow"] = overflow

    return cast(T, cls(**params))


def resolve_references(origin: Mapping[str, Any], base_path: Optional[Path] = None) -> Dict[str, Any]:
    """Resolve all reference in the object.

    :param Mapping[str, Any] origin: The object to resolve.
    :param Path|None base_path: The base path to resolve the file reference.
    :return: The resolved object.
    :rtype: Dict[str, Any]
    """

    def _resolve_references(origin: Any, base_path: Optional[Path] = None) -> Any:
        if isinstance(origin, str):
            return _resolve_reference(origin, base_path=base_path)
        if isinstance(origin, list):
            return [_resolve_references(item, base_path=base_path) for item in origin]
        if isinstance(origin, dict):
            return {key: _resolve_references(value, base_path=base_path) for key, value in origin.items()}
        return origin

    return {k: _resolve_references(v, base_path=base_path) for k, v in origin.items()}


def _resolve_reference(reference: str, base_path: Optional[Path] = None) -> Union[str, dict]:
    """
    Resolve the reference, two types are supported, env, file.
    When the string format is ${env:ENV_NAME}, the environment variable value will be returned.
    When the string format is ${file:file_path}, return the loaded json object.

    :param str reference: The reference string.
    :param Path|None base_path: The base path to resolve the file reference.
    :return: The resolved reference.
    :rtype: str | dict
    """
    pattern = r"\$\{(\w+):(.*)\}"
    match = re.match(pattern, reference)
    if match:
        reference_type, value = match.groups()
        if reference_type == "env":
            return os.environ.get(value, reference)

        if reference_type == "file":
            if not Path(value).is_absolute() and base_path:
                path = Path(base_path) / value
            else:
                path = Path(value)

            if not path.exists():
                raise PromptyException(f"Cannot find the reference file {value}.")

            with open(path, "r", encoding=DefaultOpenEncoding.READ) as f:
                if path.suffix.lower() == ".json":
                    return json.load(f)
                if path.suffix.lower() in [".yml", ".yaml"]:
                    return load_yaml(f)
                return f.read()

        # TODO ralphe: logging?
        # logger.warning(f"Unknown reference type {reference_type}, return original value {reference}.")
        return reference

    return reference


def update_dict_recursively(origin_dict: Mapping[str, Any], overwrite_dict: Mapping[str, Any]) -> Dict[str, Any]:
    updated_dict: Dict[str, Any] = {}
    for k, v in overwrite_dict.items():
        if isinstance(v, dict):
            updated_dict[k] = update_dict_recursively(origin_dict.get(k, {}), v)
        else:
            updated_dict[k] = v
    for k, v in origin_dict.items():
        if k not in updated_dict:
            updated_dict[k] = v
    return updated_dict


# endregion


# region: Jinja template rendering

VALID_ROLES = ["system", "user", "assistant", "function"]
"""Valid roles for the OpenAI Chat API"""

PROMPTY_ROLE_SEPARATOR_PATTERN = re.compile(
    r"(?i)^\s*#?\s*(" + "|".join(VALID_ROLES) + r")\s*:\s*\n", flags=re.MULTILINE
)
"""Pattern to match the role separator in a prompty template"""

MARKDOWN_IMAGE_PATTERN = re.compile(r"(?P<match>!\[[^\]]*\]\(.*?(?=\"|\))\))", flags=re.MULTILINE)
"""Pattern to match markdown syntax for embedding images such as ![alt text](url).
   This uses a 'hack' where by naming the capture group, using re.split() will cause
   the named capture group to appear in the list of split parts"""

IMAGE_URL_PARSING_PATTERN = re.compile(
    r"^!\[(?P<alt_text>[^\]]+)\]\((?P<link>(?P<scheme>[^:]+(?=:))?:?(?P<mime_type>[^;]+(?=;))?;?(?P<data>[^\)]*))\)$"
)
"""Pattern used to parse the image URL from the markdown syntax. This caputres the following groups:
    - alt_text: The alt text for the image
    - link: The full link
    - scheme: The scheme used in the link (e.g. data, http, https)
    - mime_type: The mime type of the image (only for data URLs)
    - data: The data part of the URL (only for data URLs)
"""

DEFAULT_IMAGE_MIME_TYPE: Final[str] = "image/*"
"""The mime type to use when we don't know the image type"""

FILE_EXT_TO_MIME: Final[Mapping[str, str]] = {
    ".apng": "image/apng",
    ".avif": "image/avif",
    ".bmp": "image/bmp",
    ".gif": "image/gif",
    ".heic": "image/heic",
    ".heif": "image/heif",
    ".ico": "image/vnd.microsoft.icon",
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".png": "image/png",
    ".svg": "image/svg+xml",
    ".tif": "image/tiff",
    ".tiff": "image/tiff",
    ".webp": "image/webp",
}
"""Mapping of file extensions to mime types"""


def render_jinja_template(template_str: str, *, trim_blocks=True, keep_trailing_newline=True, **kwargs) -> str:
    try:
        template = Template(template_str, trim_blocks=trim_blocks, keep_trailing_newline=keep_trailing_newline)
        return template.render(**kwargs)
    except Exception as e:  # pylint: disable=broad-except
        raise PromptyException(f"Failed to render jinja template - {type(e).__name__}: {str(e)}") from e


def build_messages(
    *, prompt: str, working_dir: Path, image_detail: str = "auto", **kwargs: Any
) -> Sequence[Mapping[str, Any]]:
    # keep_trailing_newline=True is to keep the last \n in the prompt to avoid converting "user:\t\n" to "user:".
    chat_str = render_jinja_template(prompt, trim_blocks=True, keep_trailing_newline=True, **kwargs)
    messages = _parse_chat(chat_str, working_dir, image_detail)
    return messages


def _parse_chat(chat_str: str, working_dir: Path, image_detail: str) -> Sequence[Mapping[str, Any]]:
    # openai chat api only supports VALID_ROLES as role names.
    # customer can add single # in front of role name for markdown highlight.
    # and we still support role name without # prefix for backward compatibility.

    chunks = re.split(PROMPTY_ROLE_SEPARATOR_PATTERN, chat_str)
    chat_list: List[Dict[str, Any]] = []

    for chunk in chunks:
        last_message = chat_list[-1] if len(chat_list) > 0 else None

        # =======================================================================================================
        # NOTE: The Promptflow code supported tool calls but used eval() to parse them. This is an unacceptable
        #       security risk. Since none of the current evaluators use tool calls, this functionality has been
        #       removed.
        # =======================================================================================================

        # if is_tool_chunk(last_message):
        #     parse_tools(last_message, chunk, hash2images, image_detail)
        #     continue
        # if last_message and "role" in last_message and last_message["role"] == "assistant":
        #     parsed_result = _try_parse_tool_calls(chunk)
        #     if parsed_result is not None:
        #         last_message["tool_calls"] = parsed_result
        #         continue

        if (
            last_message
            and "role" in last_message  # pylint: disable=unsupported-membership-test
            and "content" not in last_message  # pylint: disable=unsupported-membership-test
            and "tool_calls" not in last_message  # pylint: disable=unsupported-membership-test
        ):
            parsed_result = _try_parse_name_and_content(chunk)
            if parsed_result is None:
                if last_message["role"] == "function":  # pylint: disable=unsubscriptable-object
                    # "name" is required if the role is "function"
                    raise JinjaTemplateError(
                        "Failed to parse function role prompt. Please make sure the prompt follows the "
                        "format: 'name:\\nfunction_name\\ncontent:\\nfunction_content'. "
                        "'name' is required if role is function, and it should be the name of the function "
                        "whose response is in the content. May contain a-z, A-Z, 0-9, and underscores, "
                        "with a maximum length of 64 characters. See more details in "
                        "https://platform.openai.com/docs/api-reference/chat/create#chat/create-name"
                    )

                # "name" is optional for other role types.
                last_message["content"] = _to_content_str_or_list(  # pylint: disable=unsupported-assignment-operation
                    chunk, working_dir, image_detail
                )
            else:
                last_message["name"] = parsed_result[0]  # pylint: disable=unsupported-assignment-operation
                last_message["content"] = _to_content_str_or_list(  # pylint: disable=unsupported-assignment-operation
                    parsed_result[1], working_dir, image_detail
                )
        else:
            if chunk.strip() == "":
                continue
            # Check if prompt follows chat api message format and has valid role.
            # References: https://platform.openai.com/docs/api-reference/chat/create.
            role = chunk.strip().lower()
            _validate_role(role)
            new_message = {"role": role}
            chat_list.append(new_message)
    return chat_list


def _validate_role(role: str):
    if role not in VALID_ROLES:
        valid_roles_str = ", ".join(VALID_ROLES)
        error_message = (
            f"The Chat API requires a specific format for prompt definition, and the prompt should include separate "
            f"lines as role delimiters: {valid_roles_str}.\n"
            f"Current parsed role '{role}' does not meet the requirement. If you intend to use the Completion API, "
            f"please select the appropriate API type and deployment name."
        )
        raise JinjaTemplateError(message=error_message)


def _to_content_str_or_list(chat_str: str, working_dir: Path, image_detail: str) -> Union[str, List[Dict[str, Any]]]:
    chunks = [c for c in (chunk.strip() for chunk in re.split(MARKDOWN_IMAGE_PATTERN, chat_str)) if c]
    if len(chunks) <= 1:
        return chat_str.strip()

    messages: List[Dict[str, Any]] = []
    for chunk in chunks:
        if chunk.startswith("![") and chunk.endswith(")"):
            messages.append(_inline_image(chunk, working_dir, image_detail))
        else:
            messages.append({"type": "text", "text": chunk})
    return messages


def _inline_image(image: str, working_dir: Path, image_detail: str) -> Dict[str, Any]:
    """This accepts an image URL in markdown format, and parses that into a message containing the image details
    to be sent to AI service. In the case of local file images, they will be loaded and their contents encoded
    into a base 64 data URI. Internal URLs will remained untouched. It can can accept http(s), ftp(s), as well
    as data URIs.

    :param str image: The image URL in markdown format (e.g. ![alternative text](https://www.bing.com/favicon.ico))
    :param Path working_dir: The working directory to use when resolving relative file paths
    :param str image_detail: The image detail to use when sending the image to the AI service
    :return: The image message to send to the AI service
    :rtype: Mapping[str, Any]"""

    def local_to_base64(local_file: str, mime_type: Optional[str]) -> str:
        path = Path(local_file)
        if not path.is_absolute():
            path = working_dir / local_file
        if not path.exists():
            # TODO ralphe logging?
            # logger.warning(f"Cannot find the image path {image_content},
            #                  it will be regarded as {type(image_str)}.")
            raise InvalidInputError(f"Cannot find the image path '{path.as_posix()}'")

        base64_encoded = base64.b64encode(path.read_bytes()).decode("utf-8")
        if not mime_type:
            mime_type = FILE_EXT_TO_MIME.get(path.suffix.lower(), DEFAULT_IMAGE_MIME_TYPE)
        return f"data:{mime_type};base64,{base64_encoded}"

    match = re.match(IMAGE_URL_PARSING_PATTERN, image)
    if not match:
        raise InvalidInputError(f"Invalid image URL '{image}'")

    inlined_uri: str
    mime_type: Optional[str] = None

    scheme: str = (match.group("scheme") or "").strip().lower()
    if scheme in ["http", "https", "ftp", "ftps"]:
        # nothing special to do here, pass through full URI as is
        inlined_uri = (match.group("link") or "").strip()
    elif scheme == "data":
        mime_type = (match.group("mime_type") or "").strip()
        data: str = (match.group("data") or "").strip()

        # data urls may contain local paths too
        if data[:5].lower() == "path:":
            inlined_uri = local_to_base64(data[5:].strip(), mime_type)
        elif data[:6].lower() == "base64":
            # nothing special to do here, pass through full URI as is
            inlined_uri = (match.group("link") or "").strip()
        else:
            raise InvalidInputError(f"Invalid image data URL '{image}'")
    else:
        # assume it's a file path
        inlined_uri = local_to_base64((match.group("link") or "").strip(), mime_type)

    if not inlined_uri:
        raise InvalidInputError(f"Failed to determine how to inline the following image URL '{image}'")

    return {
        "type": "image_url",
        "image_url": {
            "url": inlined_uri,
            "detail": image_detail,
        },
    }


def _try_parse_name_and_content(role_prompt: str) -> Optional[Tuple[str, str]]:
    # customer can add ## in front of name/content for markdown highlight.
    # and we still support name/content without ## prefix for backward compatibility.
    # TODO ralphe: This maybe has something to do with parsing functions or tool calls but I'm not sure
    pattern = r"\n*#{0,2}\s*name\s*:\s*\n+\s*(\S+)\s*\n*#{0,2}\s*content\s*:\s*\n?(.*)"
    match = re.search(pattern, role_prompt, re.DOTALL)
    if match:
        return match.group(1), match.group(2)
    return None


# endregion


# region OpenAI connections and requests

OpenAIChatResponseType = Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]


def prepare_open_ai_request_params(
    model_config: PromptyModelConfiguration, template: Union[str, Sequence[Mapping[str, Any]]]
) -> MutableMapping[str, Any]:
    params = copy.deepcopy(model_config.parameters)
    # if isinstance(connection, AzureOpenAIConnection):
    #     params.setdefault("extra_headers", {}).update({"ms-azure-ai-promptflow-called-from": "promptflow-core"})
    params["model"] = model_config.model
    params["messages"] = template

    # NOTE:
    # - Tool calls have been disabled due to a security issue in the implementation. See comment earlier in
    #   this file for more details
    # - Removing the validation of function calls in favour of letting the service do that validation. This
    #   removes a maintenance burden from the SDK should the service definition for function calls change.

    # # functions and function_call are deprecated and are replaced by tools and tool_choice.
    # # if both are provided, tools and tool_choice are used and functions and function_call are ignored.
    # if "tools" in params:
    #     validate_tools(params["tools"])
    #     params["tool_choice"] = validate_tool_choice(params.get("tool_choice", None))
    # else:
    #     if "functions" in params:
    #         _validate_functions(params["functions"])
    #         params["function_call"] = validate_function_call(params.get("function_call", None))

    return params


async def format_llm_response(
    response: OpenAIChatResponseType,
    is_first_choice: bool,
    response_format: Optional[Mapping[str, Any]] = None,
    outputs: Optional[Mapping[str, Any]] = None,
) -> Union[OpenAIChatResponseType, AsyncGenerator[str, None], str, Mapping[str, Any]]:
    """
    Format LLM response

    If is_first_choice is false, it will directly return LLM response.
    If is_first_choice is true, behavior as blow:
        response_format: type: text
            - n: None/1/2
                Return the first choice content. Return type is string.
            - stream: True
                Return generator list of first choice content. Return type is generator[str]
        response_format: type: json_object
            - n : None/1/2
                Return json dict of the first choice. Return type is dict
            - stream: True
                Return json dict of the first choice. Return type is dict
            - outputs
                Extract corresponding output in the json dict to the first choice. Return type is dict.

    :param response: LLM response.
    :type response:
    :param is_first_choice: If true, it will return the first item in response choices, else it will return all response
    :type is_first_choice: bool
    :param response_format: An object specifying the format that the model must output.
    :type response_format: str
    :param outputs: Extract corresponding output in json format response
    :type outputs: dict
    :return: Formatted LLM response.
    :rtype: Union[str, dict, Response]
    """

    def format_choice(item: str) -> Union[str, Mapping[str, Any]]:
        # response_format is one of text or json_object.
        # https://platform.openai.com/docs/api-reference/chat/create#chat-create-response_format
        if not is_json_format:
            return item

        result_dict = json.loads(item)
        if not outputs:
            return result_dict

        # return the keys in outputs
        output_results = {}
        for key in outputs:
            if key not in result_dict:
                raise InvalidInputError(f"Cannot find '{key}' in response {list(result_dict.keys())}")
            output_results[key] = result_dict[key]
        return output_results

    async def format_stream(llm_response: AsyncStream[ChatCompletionChunk]) -> AsyncGenerator[str, None]:
        cur_index = None
        async for chunk in llm_response:
            if len(chunk.choices) > 0 and chunk.choices[0].delta.content:
                if cur_index is None:
                    cur_index = chunk.choices[0].index
                if cur_index != chunk.choices[0].index:
                    return
                yield chunk.choices[0].delta.content

    if not is_first_choice:
        return response

    is_json_format = isinstance(response_format, dict) and response_format.get("type") == "json_object"
    if isinstance(response, AsyncStream):
        if not is_json_format:
            return format_stream(llm_response=response)
        content = "".join([item async for item in format_stream(llm_response=response)])
        return format_choice(content)

    # When calling function/tool, function_call/tool_call response will be returned as a field in message,
    # so we need return message directly. Otherwise, we only return content.
    # https://platform.openai.com/docs/api-reference/chat/object#chat/object-choices
    if response.choices[0].finish_reason in ["tool_calls", "function_calls"]:
        response_content = response.model_dump()["choices"][0]["message"]
    else:
        response_content = getattr(response.choices[0].message, "content", "")
    result = format_choice(response_content)
    return result


def openai_error_retryable(
    error: OpenAIError, retry: int, entity_retry: List[int], max_entity_retries: int
) -> Tuple[bool, float]:
    """
    Determines if an OpenAI error is retryable, and optionally determines the min retry delay to use.
    If none is returned, the caller will determine the delay to use.

    :param OpenAIError error: The error to handle
    :param int retry: The current retry count (0 means we're on the first attempt and no retries have been made)
    :param List[int] entity_retry: The current retry count for the unprocessable entity failures. This should be a
        list containing only 1 element to mimic pass by reference semantics. A value of 0 means we're on the
        first attempt and no retries have been made.
    :param int max_entity_retries: The maximum number of retries to make for unprocessable entity failures
    :return: A tuple containing whether the error is retryable and the min delay to use if any
    :rtype: Tuple[bool, Optional[float]]
    """

    # Using https://platform.openai.com/docs/guides/error-codes/api-errors#python-library-error-types as a reference

    should_retry: bool
    delay: Optional[float] = None

    if isinstance(error, APIConnectionError):
        retriable_error_messages: Sequence[str] = [
            "connection aborted",
            # issue 2296
            "server disconnected without sending a response",
        ]
        should_retry = (
            isinstance(error, APITimeoutError)  # APITimeoutError is a subclass of APIConnectionError
            or str(error).lower() in retriable_error_messages
            or str(error.__cause__).lower() in retriable_error_messages
        )
    elif isinstance(error, APIStatusError):
        status_code: int = error.response.status_code
        if status_code == 422:
            # As per the original legacy code, UnprocessableEntityError (HTTP 422) should be handled differently
            # with a smaller retry count, as retrying more may not be beneficial.
            should_retry = entity_retry[0] < max_entity_retries
            entity_retry[0] += 1
        elif status_code == 429:
            # Two types, one is you are throttled and so should retry after a delay, the other is you have exceeded
            # your quota and should not retry.
            if (error.type or "").lower() == "insufficient_quota":
                should_retry = False
            else:
                should_retry = True
            should_retry = error.type != "insufficient_quota"
        else:
            should_retry = status_code >= 500

        # Use what the service tells us to use for the delay if it's provided
        if should_retry and not delay:
            delay_str = error.response.headers.get("Retry-After", None)
            if delay_str is not None:
                delay = float(delay_str)
    else:
        should_retry = False

    # Use exponential backoff for retries if the service doesn't provide a delay
    if not delay:
        delay = min(60, 2 + 2**retry)

    return (should_retry, delay)


# endregion



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import os
import types
from typing import Optional, Type, Union

from azure.ai.evaluation._legacy._adapters._constants import PF_FLOW_ENTRY_IN_TMP, PF_FLOW_META_LOAD_IN_SUBPROCESS
from azure.ai.evaluation._legacy._adapters.utils import ClientUserAgentUtil
from azure.ai.evaluation._legacy._adapters.tracing import inject_openai_api, recover_openai_api
from azure.ai.evaluation._legacy._batch_engine._openai_injector import (
    inject_openai_api as ported_inject_openai_api,
    recover_openai_api as ported_recover_openai_api,
)

from azure.ai.evaluation._constants import (
    OTEL_EXPORTER_OTLP_TRACES_TIMEOUT,
    OTEL_EXPORTER_OTLP_TRACES_TIMEOUT_DEFAULT,
    PF_BATCH_TIMEOUT_SEC,
    PF_BATCH_TIMEOUT_SEC_DEFAULT,
    PF_DISABLE_TRACING,
)

from ..._user_agent import USER_AGENT
from .._utils import set_event_loop_policy
from .batch_clients import BatchClient
from ._run_submitter_client import RunSubmitterClient
from .code_client import CodeClient
from .proxy_client import ProxyClient


class EvalRunContext:
    """Context manager for eval batch run.

    :param client: The client to run in the context.
    :type client: Union[
        ~azure.ai.evaluation._evaluate._batch_run.code_client.CodeClient,
        ~azure.ai.evaluation._evaluate._batch_run.proxy_client.ProxyClient
    ]
    """

    def __init__(self, client: BatchClient) -> None:
        self.client = client
        self._is_batch_timeout_set_by_system = False
        self._is_otel_timeout_set_by_system = False
        self._original_cwd = os.getcwd()

    def __enter__(self) -> None:
        # Preserve current working directory, as PF may change it without restoring it afterward
        self._original_cwd = os.getcwd()

        if isinstance(self.client, CodeClient):
            ClientUserAgentUtil.append_user_agent(USER_AGENT)
            inject_openai_api()

        if isinstance(self.client, ProxyClient):
            os.environ[PF_FLOW_ENTRY_IN_TMP] = "true"
            os.environ[PF_FLOW_META_LOAD_IN_SUBPROCESS] = "false"
            os.environ[PF_DISABLE_TRACING] = "true"

            if os.environ.get(PF_BATCH_TIMEOUT_SEC) is None:
                os.environ[PF_BATCH_TIMEOUT_SEC] = str(PF_BATCH_TIMEOUT_SEC_DEFAULT)
                self._is_batch_timeout_set_by_system = True

            # For dealing with the timeout issue of OpenTelemetry exporter when multiple evaluators are running
            if os.environ.get(OTEL_EXPORTER_OTLP_TRACES_TIMEOUT) is None:
                os.environ[OTEL_EXPORTER_OTLP_TRACES_TIMEOUT] = str(OTEL_EXPORTER_OTLP_TRACES_TIMEOUT_DEFAULT)
                self._is_otel_timeout_set_by_system = True

            # For addressing the issue of asyncio event loop closed on Windows
            set_event_loop_policy()

        if isinstance(self.client, RunSubmitterClient):
            set_event_loop_policy()
            ported_inject_openai_api()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_tb: Optional[types.TracebackType],
    ) -> None:
        os.chdir(self._original_cwd)

        if isinstance(self.client, CodeClient):
            recover_openai_api()

        if isinstance(self.client, ProxyClient):
            os.environ.pop(PF_FLOW_ENTRY_IN_TMP, None)
            os.environ.pop(PF_FLOW_META_LOAD_IN_SUBPROCESS, None)
            os.environ.pop(PF_DISABLE_TRACING, None)

            if self._is_batch_timeout_set_by_system:
                os.environ.pop(PF_BATCH_TIMEOUT_SEC, None)
                self._is_batch_timeout_set_by_system = False

            if self._is_otel_timeout_set_by_system:
                os.environ.pop(OTEL_EXPORTER_OTLP_TRACES_TIMEOUT, None)
                self._is_otel_timeout_set_by_system = False

        if isinstance(self.client, RunSubmitterClient):
            ported_recover_openai_api()



import json
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed

from azure.ai.projects import __version__ as projects_version
from azure.ai.projects import AIProjectClient

from typing import List, Union

from azure.ai.evaluation._common._experimental import experimental
from packaging.version import Version

# Constants.
from ._models import _USER, _AGENT, _TOOL, _TOOL_CALL, _TOOL_CALLS, _FUNCTION, _BUILT_IN_DESCRIPTIONS, _BUILT_IN_PARAMS

# Message instances.
from ._models import Message, SystemMessage, UserMessage, AssistantMessage, ToolCall

# Intermediate definitions to hold results.
from ._models import ToolDefinition, EvaluatorData

# Utilities.
from ._models import break_tool_call_into_messages, convert_message


@experimental
class AIAgentConverter:
    """
    A converter for AI agent data.  Data retrieval classes handle getting agent data depending on
    agent version.

    :param project_client: The AI project client used for API interactions.
    :type project_client: AIProjectClient
    """

    # Maximum number of workers allowed to make API calls at the same time.
    _MAX_WORKERS = 10

    def __init__(self, project_client: AIProjectClient):
        """
        Initializes the AIAgentConverter with the given AI project client.

        :param project_client: The AI project client used for API interactions.
        :type project_client: AIProjectClient
        """
        self.project_client = project_client
        self._data_retriever = AIAgentConverter._get_data_retriever(project_client=project_client)

    @staticmethod
    def _get_data_retriever(project_client: AIProjectClient):
        if project_client is None:
            return None
        if Version(projects_version) > Version("1.0.0b10"):
            return FDPAgentDataRetriever(project_client=project_client)
        else:
            return LegacyAgentDataRetriever(project_client=project_client)

    def _list_tool_calls_chronological(self, thread_id: str, run_id: str) -> List[ToolCall]:
        """
        Lists tool calls in chronological order for a given thread and run.

        :param thread_id: The ID of the thread.
        :type thread_id: str
        :param run_id: The ID of the run.
        :type run_id: str
        :return: A list of tool calls in chronological order.
        :rtype: List[ToolCall]
        """
        # This is the other API request that we need to make to AI service, such that we can get the details about
        # the tool calls and results. Since the list is given in reverse chronological order, we need to reverse it.
        run_steps_chronological = self._data_retriever._list_run_steps_chronological(thread_id=thread_id, run_id=run_id)

        # Let's accumulate the function calls in chronological order. Function calls
        tool_calls_chronological: List[ToolCall] = []
        for run_step_chronological in run_steps_chronological:
            if run_step_chronological.type != _TOOL_CALLS:
                continue
            step_details: object = run_step_chronological.step_details
            if step_details.type != _TOOL_CALLS:
                continue
            if len(step_details.tool_calls) < 1:
                continue
            for tool_call in step_details.tool_calls:
                # We need to add the tool call and the result as two separate messages.
                tool_calls_chronological.append(
                    ToolCall(
                        created=run_step_chronological.created_at,
                        completed=run_step_chronological.completed_at,
                        details=tool_call,
                    )
                )

        return tool_calls_chronological

    @staticmethod
    def _extract_function_tool_definitions(thread_run: object) -> List[ToolDefinition]:
        """
        Extracts tool definitions from a thread run.

        :param thread_run: The thread run containing tool definitions.
        :type thread_run: object
        :return: A list of tool definitions extracted from the thread run.
        :rtype: List[ToolDefinition]
        """
        final_tools: List[ToolDefinition] = []
        for tool in thread_run.tools:
            # Here we handle the custom functions and create tool definitions out of them.
            if tool.type == _FUNCTION:
                tool_function: FunctionDefinition = tool.function
                parameters = tool_function.parameters

                # The target schema doesn't support required fields, so we omit it for now.
                if "required" in parameters:
                    del parameters["required"]

                final_tools.append(
                    ToolDefinition(
                        type="function",
                        name=tool_function.name,
                        description=tool_function.description,
                        parameters=parameters,
                    )
                )
            else:
                # Add limited support for built-in tools. Descriptions and parameters
                # are not published, but we'll include placeholders.
                if tool.type in _BUILT_IN_DESCRIPTIONS and tool.type in _BUILT_IN_PARAMS:
                    final_tools.append(
                        ToolDefinition(
                            type=tool.type,
                            name=tool.type,
                            description=_BUILT_IN_DESCRIPTIONS[tool.type],
                            parameters=_BUILT_IN_PARAMS[tool.type],
                        )
                    )
        return final_tools

    @staticmethod
    def _break_into_query_responses(messages: List[Message], run_id: str) -> (List[Message], List[Message]):
        """
        Breaks a list of messages into query and response messages based on the run ID.

        :param messages: The list of messages to be broken into query and response.
        :type messages: List[Message]
        :param run_id: The ID of the run to distinguish response messages.
        :type run_id: str
        :return: A tuple containing two lists - the first list contains query messages, and the second list contains response messages.
        :rtype: (List[Message], List[Message])
        """
        query: List[Message] = [what for what in messages if what.run_id != run_id]
        responses: List[Message] = [what for what in messages if what.run_id == run_id]
        return query, responses

    @staticmethod
    def _filter_run_ids_up_to_run_id(run_ids: List[str], run_id: str, include_run_id: bool = True) -> List[str]:
        """
        Filters run IDs up to a specific run ID.

        This method processes a list of run IDs and filters out run IDs that come after the specified run ID.
        It ensures that only run IDs up to and including the specified run ID are included in the result.

        :param run_ids: The list of run IDs in chronological order.
        :type run_ids: List[str]
        :param run_id: The ID of the run to filter messages up to.
        :type run_id: str
        :return: The filtered list of run IDs up to the specified run ID.
        :rtype: List[str]
        """
        for index, single_run_id in enumerate(run_ids):
            # Since this is the conversation of the entire thread and we are interested in a given run, we need to
            # filter out the messages that came after the run.
            if single_run_id == run_id:
                if include_run_id:
                    return run_ids[: index + 1]
                return run_ids[:index]

        # If we didn't find the run_id, we return an empty list.
        return []

    @staticmethod
    def _filter_messages_up_to_run_id(
        chronological_messages, run_id: str, include_run_id: bool = True
    ) -> List[Message]:
        """
        Filters messages up to a specific run ID.

        This method processes a list of messages in chronological order and filters out messages that come after the specified run ID.
        It ensures that only messages up to and including the specified run ID are included in the result.

        :param chronological_messages: The list of messages in chronological order.
        :type chronological_messages: List[Message]
        :param run_id: The ID of the run to filter messages up to.
        :type run_id: str
        :return: The filtered list of messages up to the specified run ID.
        :rtype: List[Message]
        """
        filtered_messages = []
        in_my_current_run = False
        for single_turn in chronological_messages:
            # Since this is the conversation of the entire thread and we are interested in a given run, we need to
            # filter out the messages that came after the run.
            if single_turn.run_id is not None:
                if single_turn.run_id == run_id:
                    in_my_current_run = True

                    # If we entered our current run and its the run that we have requested to filter up to, but
                    # not including, we can break out of the loop.
                    if not include_run_id:
                        break

            # Then, if we think that we are currently in our run and we have a message that is not from our run,
            # it means that we have left our run.
            if in_my_current_run and single_turn.run_id != run_id:
                break

            # We're good to add it.
            filtered_messages.append(single_turn)

        return filtered_messages

    @staticmethod
    def _extract_typed_messages(ai_services_messages) -> List[Message]:
        """
        Extracts and converts AI service messages to a list of typed Message objects.

        This method processes a list of messages from the AI service, converting them into
        appropriate Message subclass instances (UserMessage, AssistantMessage) based on their role.
        It filters out messages without content and handles different message roles accordingly.

        :param ai_services_messages: A list of messages from the AI service.
        :type ai_services_messages: _models.OpenAIPageableListOfThreadMessage (some internal type from ai projects)
        :return: A list of typed Message objects.
        :rtype: List[Message]
        """
        # We will collect messages in this accumulator.
        final_messages: List[Message] = []

        # Each visible message in the conversation is a message from the user or the assistant, we collect
        # both the text and timestamp, so we can recreate the chronological order.
        for single_turn in ai_services_messages:
            # This shouldn't really happen, ever. What's the point of a message without content? But to avoid a nasty
            # crash on one of the historical messages, let's check for it and bail out from this iteration.
            if len(single_turn.content) < 1:
                continue

            # Build the content of the text message.
            content = {
                "type": "text",
                "text": single_turn.content[0].text.value,
            }

            # If we have a user message, then we save it as such and since it's a human message, there is no
            # run_id associated with it.
            if single_turn.role == _USER:
                final_messages.append(UserMessage(content=[content], createdAt=single_turn.created_at))
                continue

            # In this case, we have an assistant message. Unfortunately, this would only have the user-facing
            # agent's response, without any details on what tool was called, with what parameters, and what
            # the result was. That will be added later in the method.
            if single_turn.role == _AGENT:
                # We are required to put the run_id in the assistant message.
                final_messages.append(
                    AssistantMessage(content=[content], run_id=single_turn.run_id, createdAt=single_turn.created_at)
                )
                continue

        return final_messages

    def _fetch_tool_calls(self, thread_id: str, run_id: str) -> List[Message]:
        """
        Fetches tool calls for a given thread and run, and converts them into messages.

        This method retrieves tool calls for a specified thread and run, converts them into messages using the
        `break_tool_call_into_messages` utility function, and returns the list of messages.

        :param thread_id: The ID of the thread.
        :type thread_id: str
        :param run_id: The ID of the run.
        :type run_id: str
        :return: A list of messages converted from tool calls.
        :rtype: List[Message]
        """
        tool_calls: List[Message] = []
        for chrono_tool_call in self._list_tool_calls_chronological(thread_id, run_id):
            tool_calls.extend(break_tool_call_into_messages(chrono_tool_call, run_id))
        return tool_calls

    def _retrieve_tool_calls_up_to_including_run_id(
        self, thread_id: str, run_id: str, exclude_tool_calls_previous_runs: bool = False
    ) -> List[Message]:
        """
        Converts tool calls to messages for a given thread and run.

        This method retrieves tool calls for a specified thread and run, converts them into messages,
        and optionally includes tool calls from previous runs.

        :param thread_id: The ID of the thread.
        :type thread_id: str
        :param run_id: The ID of the run.
        :type run_id: str
        :param exclude_tool_calls_previous_runs: Whether to exclude tool calls from previous runs in the conversion. Default is False.
        :type exclude_tool_calls_previous_runs: bool
        :return: A list of messages converted from tool calls.
        :rtype: List[Message]
        """
        to_return: List[Message] = []

        # Add all the tool calls and results of this run as messages.
        for tool_call in self._list_tool_calls_chronological(thread_id, run_id):
            # We need to add the tool call and the result as two separate messages.
            to_return.extend(break_tool_call_into_messages(tool_call, run_id))

        # We also request to add all the tool calls and results of the previous runs into the chat history. This is
        # a bit of an expensive operation, but the requirement is to support this functionality, even at the penalty
        # in latency in performance. New agents api is to include these details cheaply through a single API call in
        # list_messages, but until that is available, we need to do this. User can also opt-out of this functionality
        # by setting the exclude_tool_calls_previous_runs flag to True.
        if not exclude_tool_calls_previous_runs:
            # These are all the assistant (any number) in the thread.
            # We set the include_run_id to False, since we don't want to include the current run's tool calls, which
            # are already included in the previous step.
            run_ids_up_to_run_id = AIAgentConverter._filter_run_ids_up_to_run_id(
                self._data_retriever._list_run_ids_chronological(thread_id), run_id, include_run_id=False
            )

            # Since each _list_tool_calls_chronological call is expensive, we can use a thread pool to speed
            # up the process by parallelizing the AI Services API requests.
            with ThreadPoolExecutor(max_workers=self._MAX_WORKERS) as executor:
                futures = {
                    executor.submit(self._fetch_tool_calls, thread_id, run_id): run_id
                    for run_id in run_ids_up_to_run_id
                }
                for future in as_completed(futures):
                    to_return.extend(future.result())

        return to_return

    def _retrieve_all_tool_calls(self, thread_id: str, run_ids: List[str]) -> List[Message]:
        """
        Converts all tool calls to messages for a given thread and list of run IDs.

        This method retrieves tool calls for a specified thread and list of run IDs, converts them into messages,
        and returns the list of messages.

        :param thread_id: The ID of the thread.
        :type thread_id: str
        :param run_ids: The list of run IDs.
        :type run_ids: List[str]
        :return: A list of messages converted from tool calls.
        :rtype: List[Message]
        """
        to_return: List[Message] = []

        with ThreadPoolExecutor(max_workers=self._MAX_WORKERS) as executor:
            futures = {executor.submit(self._fetch_tool_calls, thread_id, run_id): run_id for run_id in run_ids}
            for future in as_completed(futures):
                to_return.extend(future.result())

        return to_return

    @staticmethod
    def _is_agent_tool_call(message: Message) -> bool:
        """
        Determines if a message is an agent tool call.

        :param message: The message to be checked.
        :type message: Message
        :return: True if the message is an agent tool call, False otherwise.
        :rtype: bool
        """
        return (
            message.role == _AGENT  # Any other agent that this run's.
            and isinstance(message.content, list)  # Content is of expected type.
            and len(message.content) > 0  # There are messages/calls/results present.
            and "type" in message.content[0]  # Being safe here.
            and message.content[0]["type"] == _TOOL_CALL  # Not interested in assistant's toolcalls.
        )

    @staticmethod
    def _sort_messages(messages: List[Message]) -> List[Message]:
        """
        Sorts a list of messages, placing messages with `createdAt` set to None at the beginning.

        :param messages: The list of messages to be sorted.
        :type messages: List[Message]
        :return: The sorted list of messages.
        :rtype: List[Message]
        """
        # Separate messages with createdAt set to None
        none_created_at = [message for message in messages if message.createdAt is None]

        # Filter out messages with createdAt set to None and sort the remaining messages
        sorted_messages = sorted(
            [message for message in messages if message.createdAt is not None],
            key=lambda x: (x.createdAt, x.role == _AGENT),
        )

        # Combine the lists, placing messages with None createdAt at the beginning
        return none_created_at + sorted_messages

    def convert(self, thread_id: str, run_id: str, exclude_tool_calls_previous_runs: bool = False) -> dict:
        """
        Converts the agent run to a format suitable for the OpenAI API.

        :param thread_id: The ID of the thread.
        :type thread_id: str
        :param run_id: The ID of the run.
        :type run_id: str
        :param exclude_tool_calls_previous_runs: Whether to exclude tool calls from previous runs in the conversion.
        :type exclude_tool_calls_previous_runs: bool
        :return: The converted data in dictionary format.
        :rtype: dict
        """
        # Make the API call once and reuse the result.
        thread_run: object = self._data_retriever._get_run(thread_id=thread_id, run_id=run_id)

        # Walk through the "user-facing" conversation history and start adding messages.
        chronological_conversation = self._data_retriever._list_messages_chronological(thread_id)

        # Since this is Xth run of out possibly N runs, we are only interested is messages that are before the run X.
        chrono_until_run_id = AIAgentConverter._filter_messages_up_to_run_id(chronological_conversation, run_id)

        # Messages are now still in hidden AI services' type, so to get finer control over our typing, we need to
        # convert the message to a friendly schema.
        final_messages = AIAgentConverter._extract_typed_messages(chrono_until_run_id)

        # Third, add all the tool calls and results as messages.
        final_messages.extend(
            self._retrieve_tool_calls_up_to_including_run_id(thread_id, run_id, exclude_tool_calls_previous_runs)
        )

        # All of our final messages have to be in chronological order. We use a secondary sorting key,
        # since the tool_result and assistant events would come with the same timestamp, so we need to
        # sort them by role, such that the assistant's message would come after the tool result it's sending.
        final_messages = AIAgentConverter._sort_messages(final_messages)

        # Finally, we want to force the system message to be the first one in the list.
        # First, we need to create the first system message of the thread.
        instructions = thread_run.instructions
        if instructions:
            # The system message will have a string content.
            final_messages.insert(0, SystemMessage(content=instructions))

        # We need to collect all the messages that are not the current run's response.
        query, responses = AIAgentConverter._break_into_query_responses(final_messages, run_id)

        # Collect it into the final result and dump it to JSON.
        final_result = EvaluatorData(
            query=query,
            response=responses,
            tool_definitions=AIAgentConverter._extract_function_tool_definitions(thread_run),
        )

        return json.loads(final_result.to_json())

    def _prepare_single_thread_evaluation_data(self, thread_id: str, filename: str = None) -> List[dict]:
        """
        Prepares evaluation data for a given thread and optionally writes it to a file.

        This method retrieves all run IDs and messages for the specified thread, processes them to create evaluation data,
        and optionally writes the evaluation data to a JSONL file. The evaluation data includes query and response messages
        as well as tool definitions.

        :param thread_id: The ID of the thread.
        :type thread_id: str
        :param filename: The name of the file to write the evaluation data to. If None, the data is not written to a file.
        :type filename: str, optional
        :return: A list of evaluation data dictionaries.
        :rtype: List[dict]
        """
        list_of_run_evaluations: List[dict] = []

        # These are all the run IDs.
        run_ids = self._data_retriever._list_run_ids_chronological(thread_id)

        # If there were no messages in the thread, we can return an empty list.
        if len(run_ids) < 1:
            return list_of_run_evaluations

        # These are all the messages.
        chronological_conversation = self._data_retriever._list_messages_chronological(thread_id)

        # If there are no messages in the thread, we can return an empty list.
        if len(chronological_conversation) < 1:
            return list_of_run_evaluations

        # These are all the tool calls.
        all_sorted_tool_calls = AIAgentConverter._sort_messages(self._retrieve_all_tool_calls(thread_id, run_ids))

        # The last run should have all the tool definitions.
        thread_run = self._data_retriever._get_run(thread_id=thread_id, run_id=run_ids[-1])
        instructions = thread_run.instructions

        # So then we can get the tool definitions.
        tool_definitions = AIAgentConverter._extract_function_tool_definitions(thread_run)

        # Now, we create a new evaluator object for each run.
        for run_id in run_ids:
            # We need to filter out the messages that are not from the current run.
            simple_messages = AIAgentConverter._filter_messages_up_to_run_id(chronological_conversation, run_id)

            # Now we need to convert from OpenAI's general ThreadMessage model into our Azure Agents models.
            typed_simple_messages = AIAgentConverter._extract_typed_messages(simple_messages)

            # We also need to filter out the tool calls that are not from the current run.
            sorted_tool_calls = AIAgentConverter._filter_messages_up_to_run_id(all_sorted_tool_calls, run_id)

            # Build the big list.
            this_runs_messages = []
            this_runs_messages.extend(typed_simple_messages)
            this_runs_messages.extend(sorted_tool_calls)

            # Sort it, so it looks nicely in chronological order.
            this_runs_messages = AIAgentConverter._sort_messages(this_runs_messages)

            # If we have a system message, we need to put it at the top of the list.
            if instructions:
                # The system message will have a string content.
                this_runs_messages.insert(0, SystemMessage(content=instructions))

            # Since now we have the messages in the expected order, we need to break them into the query and
            # responses.
            query, responses = AIAgentConverter._break_into_query_responses(this_runs_messages, run_id)

            # Finally, let's pack it up into the final result.
            final_result = EvaluatorData(
                query=query,
                response=responses,
                tool_definitions=tool_definitions,
            )

            # Add it to the list of evaluations.
            list_of_run_evaluations.append(json.loads(final_result.to_json()))

        # So, if we have the filename, we can write it to the file, which is expected to be a JSONL file.
        if filename:
            with open(filename, mode="a", encoding="utf-8") as file:
                for evaluation in list_of_run_evaluations:
                    file.write(json.dumps(evaluation) + "\n")

        # We always return the list of evaluations, even if we didn't or did write it to a file.
        return list_of_run_evaluations

    def prepare_evaluation_data(self, thread_ids=Union[str, List[str]], filename: str = None) -> List[dict]:
        """
        Prepares evaluation data for a given thread or list of threads and optionally writes it to a file.

        This method retrieves all run IDs and messages for the specified thread(s), processes them to create evaluation data,
        and optionally writes the evaluation data to a JSONL file. The evaluation data includes query and response messages
        as well as tool definitions.

        :param thread_ids: The ID(s) of the thread(s). Can be a single thread ID or a list of thread IDs.
        :type thread_ids: Union[str, List[str]]
        :param filename: The name of the file to write the evaluation data to. If None, the data is not written to a file.
        :type filename: str, optional
        :return: A list of evaluation data dictionaries.
        :rtype: List[dict]
        """
        # Single instance, pretty much the same as the list.
        if isinstance(thread_ids, str):
            return self._prepare_single_thread_evaluation_data(thread_id=thread_ids, filename=filename)

        evaluations = []
        with ThreadPoolExecutor(max_workers=self._MAX_WORKERS) as executor:
            # We override the filename, because we don't want to write the file for each thread, having to handle
            # threading issues and file being opened from multiple threads, instead, we just want to write it once
            # at the end.
            futures = {
                executor.submit(self._prepare_single_thread_evaluation_data, str(thread_id), None): thread_id
                for thread_id in thread_ids
            }
            for future in as_completed(futures):
                evaluations.extend(future.result())

        # So, if we have the filename, we can write it to the file, which is expected to be a JSONL file.
        if filename:
            with open(filename, mode="a", encoding="utf-8") as file:
                for evaluation in evaluations:
                    file.write(json.dumps(evaluation) + "\n")

        return evaluations

    @staticmethod
    def _run_ids_from_conversation(conversation: dict) -> List[str]:
        """
        Extracts a list of unique run IDs from a conversation dictionary.

        :param conversation: The conversation dictionary containing messages.
        :type conversation: dict
        :return: A list of unique run IDs in the order they appear.
        :rtype: List[str]
        """
        if not isinstance(conversation, dict) or "messages" not in conversation:
            return []
        run_ids_with_repetitions = [message["run_id"] for message in conversation["messages"] if "run_id" in message]
        # Removes duplicates, requires Python 3.7+ to ensure order is preserved
        run_ids = list(dict.fromkeys(run_ids_with_repetitions))
        return run_ids

    @staticmethod
    def _convert_from_conversation(
        conversation: dict, run_id: str, exclude_tool_calls_previous_runs: bool = False
    ) -> dict:
        """
        Converts the agent run from a conversation dictionary object (a loaded thread) to a format suitable for the OpenAI API.

        :param conversation: The conversation dictionary object.
            The expected schema for the conversation dictionary is as follows:
            {
                "messages": [
                    {
                        "role": str,  # The role of the message sender, e.g., "user", "assistant", "tool".
                        "content": list,  # A list of content dictionaries.
                        "run_id": str,  # The ID of the run associated with the message.
                        "createdAt": str,  # The timestamp when the message was created.
                        ...
                    },
                    ...
                ],
                "tools": [
                    {
                        "name": str,  # The name of the tool.
                        "description": str,  # The description of the tool.
                        "parameters": dict  # The parameters for the tool.
                    },
                    ...
                ]
            }
        :type conversation: dict
        :param run_id: The ID of the run.
        :type run_id: str
        :param exclude_tool_calls_previous_runs: Whether to exclude tool calls from previous runs in the conversion.
        :type exclude_tool_calls_previous_runs: bool
        :return: The converted data in dictionary format serialized as string.
        :rtype: dict
        """
        # We need to type our messages to the correct type, so we can sliced and dice the way we like it.
        messages: List[dict] = conversation.get("messages", [])
        converted_messages: List[Message] = [convert_message(message) for message in messages]

        # Accumulate the messages in the correct order, but only up to the run_id.
        final_messages: List[Message] = []
        for converted_message in AIAgentConverter._filter_messages_up_to_run_id(converted_messages, run_id):
            # By default, we want to add all the messages, even if we are on the 10th run of the thread, we want to know
            # what the assistant said, what the assistant called, and what the result was.
            if exclude_tool_calls_previous_runs:
                # We would not be interested in tool call messages in the query, unless it's the current run id.
                if converted_message.run_id != run_id:
                    # Anything with tool, we can throw out, since we don't care about the tooling of possibly other agents
                    # that came before the run we're interested in.
                    if converted_message.role == _TOOL:
                        continue

                    # We also don't want anything that is an assistant calling a tool.
                    if AIAgentConverter._is_agent_tool_call(converted_message):
                        continue

            # We're good to add it.
            final_messages.append(converted_message)

        # Just in case, sort them all out by putting the messages without createdAt, like SystemMessage's at the
        # top of the list, so they appear first.
        final_messages = AIAgentConverter._sort_messages(final_messages)

        # Create the tool definitions.
        tools = conversation.get("tools", [])
        tool_definitions = [
            ToolDefinition(name=tool["name"], description=tool.get("description"), parameters=tool["parameters"])
            for tool in tools
        ]

        # Separate into the chat history, with all other user-assistant messages, and the assistant's response, where
        # the latter would include
        query, responses = AIAgentConverter._break_into_query_responses(final_messages, run_id)

        # Create the final result
        final_result = EvaluatorData(query=query, response=responses, tool_definitions=tool_definitions)

        return json.loads(final_result.to_json())

    @staticmethod
    def _convert_from_file(filename: str, run_id: str) -> dict:
        """
        Converts the agent run from a JSON file to a format suitable for the OpenAI API, the JSON file being a thread.

        :param filename: The path to the JSON file.
            The expected schema for the JSON file is as follows:
            {
                "messages": [
                    {
                        "role": str,  # The role of the message sender, e.g., "user", "assistant", "tool".
                        "content": list,  # A list of content dictionaries.
                        "run_id": str,  # The ID of the run associated with the message.
                        "createdAt": str,  # The timestamp when the message was created.
                        ...
                    },
                    ...
                ],
                "tools": [
                    {
                        "name": str,  # The name of the tool.
                        "description": str,  # The description of the tool.
                        "parameters": dict  # The parameters for the tool.
                    },
                    ...
                ]
            }
        :type filename: str
        :param run_id: The ID of the run.
        :type run_id: str
        :return: The converted data in dictionary format serialized as string.
        :rtype: dict
        """

        with open(filename, mode="r", encoding="utf-8") as file:
            data = json.load(file)

        return AIAgentConverter._convert_from_conversation(data, run_id)

@experimental
class AIAgentDataRetriever:
    # Maximum items to fetch in a single AI Services API call (imposed by the service).
    _AI_SERVICES_API_MAX_LIMIT = 100

    def __init__(self, project_client: AIProjectClient):
        """
        Initializes the AIAgentDataRetriever with the given AI project client.

        :param project_client: The AI project client used for API interactions.
        :type project_client: AIProjectClient
        """
        self.project_client = project_client

    @abstractmethod
    def _get_run(self, thread_id: str, run_id: str):
        pass

    @abstractmethod
    def _list_messages_chronological(self, thread_id: str):
        pass

    @abstractmethod
    def _list_run_steps_chronological(self, thread_id: str, run_id: str):
        pass

    @abstractmethod
    def _list_run_ids_chronological(self, thread_id: str) -> List[str]:
        pass

@experimental
class LegacyAgentDataRetriever(AIAgentDataRetriever):

    def __init__(self, **kwargs):
        super(LegacyAgentDataRetriever, self).__init__(**kwargs)

    def _list_messages_chronological(self, thread_id: str):
        """
        Lists messages in chronological order for a given thread.

        :param thread_id: The ID of the thread.
        :type thread_id: str
        :return: A list of messages in chronological order.
        """
        to_return = []

        has_more = True
        after = None
        while has_more:
            messages = self.project_client.agents.list_messages(
            thread_id=thread_id, limit=self._AI_SERVICES_API_MAX_LIMIT, order="asc", after=after)
            has_more = messages.has_more
            after = messages.last_id
            if messages.data:
                # We need to add the messages to the accumulator.
                to_return.extend(messages.data)

        return to_return

    def _list_run_steps_chronological(self, thread_id: str, run_id: str):
        run_steps_chronological: List[object] = []
        has_more = True
        after = None
        while has_more:
            run_steps = self.project_client.agents.list_run_steps(
                thread_id=thread_id,
                run_id=run_id,
                limit=self._AI_SERVICES_API_MAX_LIMIT,
                order="asc",
                after=after,
            )
            has_more = run_steps.has_more
            after = run_steps.last_id
            if run_steps.data:
                # We need to add the run steps to the accumulator.
                run_steps_chronological.extend(run_steps.data)
        return run_steps_chronological

    def _list_run_ids_chronological(self, thread_id: str) -> List[str]:
        """
        Lists run IDs in chronological order for a given thread.

        :param thread_id: The ID of the thread.
        :type thread_id: str
        :return: A list of run IDs in chronological order.
        :rtype: List[str]
        """
        runs = self.project_client.agents.list_runs(thread_id=thread_id, order="asc")
        run_ids = [run["id"] for run in runs["data"]]
        return run_ids

    def _get_run(self, thread_id: str, run_id: str):
        return self.project_client.agents.get_run(thread_id=thread_id, run_id=run_id)

@experimental
class FDPAgentDataRetriever(AIAgentDataRetriever):

    def __init__(self, **kwargs):
        super(FDPAgentDataRetriever, self).__init__(**kwargs)

    def _list_messages_chronological(self, thread_id: str):
        """
        Lists messages in chronological order for a given thread.

        :param thread_id: The ID of the thread.
        :type thread_id: str
        :return: A list of messages in chronological order.
        """
        message_iter = self.project_client.agents.messages.list(
            thread_id=thread_id, limit=self._AI_SERVICES_API_MAX_LIMIT, order="asc"
        )
        return [message for message in message_iter]

    def _list_run_steps_chronological(self, thread_id: str, run_id: str):

        return  self.project_client.agents.run_steps.list(
                thread_id=thread_id,
                run_id=run_id,
                limit=self._AI_SERVICES_API_MAX_LIMIT,
                order="asc"
            )

    def _list_run_ids_chronological(self, thread_id: str) -> List[str]:
        runs = self.project_client.agents.runs.list(thread_id=thread_id, order="asc")
        return [run.id for run in runs]

    def _get_run(self, thread_id: str, run_id: str):
        return self.project_client.agents.runs.get(thread_id=thread_id, run_id=run_id)


# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Final, Optional
from typing_extensions import TypeAlias


try:
    from promptflow._utils.user_agent_utils import ClientUserAgentUtil as _ClientUserAgentUtil
    from promptflow._utils.async_utils import async_run_allowing_running_loop as _async_run_allowing_running_loop
    from promptflow._cli._utils import get_workspace_triad_from_local as _get_workspace_triad_from_local
except ImportError:
    from azure.ai.evaluation._legacy._batch_engine._utils_deprecated import (
        async_run_allowing_running_loop as _async_run_allowing_running_loop,
    )
    from azure.ai.evaluation._evaluate._utils import AzureMLWorkspace

    class _ClientUserAgentUtil:
        @staticmethod
        def append_user_agent(user_agent: Optional[str]):
            # TODO ralphe: implement?
            pass

    def _get_workspace_triad_from_local() -> AzureMLWorkspace:
        return AzureMLWorkspace("", "", "")


ClientUserAgentUtil: TypeAlias = _ClientUserAgentUtil
async_run_allowing_running_loop: Final = _async_run_allowing_running_loop
get_workspace_triad_from_local: Final = _get_workspace_triad_from_local



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import math
from typing import List, Callable, Any

from azure.ai.evaluation._exceptions import EvaluationException, ErrorBlame, ErrorCategory, ErrorTarget


def list_sum(lst: List[float]) -> float:
    """Given a list of floats, return the sum of the values.

    :param lst: A list of floats.
    :type lst: List[float]
    :return: The sum of the values in the list.
    :rtype: float
    """

    return sum(lst)


def list_mean(lst: List[float]) -> float:
    """Given a list of floats, calculate the mean of the values.

    :param lst: A list of floats.
    :type lst: List[float]
    :return: The mean of the values in the list.
    :rtype: float
    """

    return list_sum(lst) / len(lst)


def list_mean_nan_safe(lst: List[float]) -> float:
    """Given a list of floats, remove all nan or None values, then calculate the mean of the remaining values.

    :param lst: A list of floats.
    :type lst: List[float]
    :return: The mean of the values in the list.
    :rtype: float
    """

    msg = "All score values are NaN. The mean cannot be calculated."
    if all(math.isnan(l) for l in lst):
        raise EvaluationException(
            message=msg,
            internal_message=msg,
            blame=ErrorBlame.USER_ERROR,
            category=ErrorCategory.INVALID_VALUE,
            target=ErrorTarget.CONVERSATION,
        )
    return list_mean([l for l in lst if not is_none_or_nan(l)])


def apply_transform_nan_safe(lst: List[float], transform_fn: Callable[[float], Any]) -> List[Any]:
    """Given a list of floats, remove all nan values, then apply the inputted transform function
    to the remaining values, and return the resulting list of outputted values.

    :param lst: A list of floats.
    :type lst: List[float]
    :param transform_fn: A function that produces something when applied to a float.
    :type transform_fn: Callable[[float], Any]
    :return: A list of the transformed values.
    :rtype: List[Any]
    """

    msg = "All score values are NaN. The mean cannot be calculated."
    if all(math.isnan(l) for l in lst):
        raise EvaluationException(
            message=msg,
            internal_message=msg,
            blame=ErrorBlame.USER_ERROR,
            category=ErrorCategory.INVALID_VALUE,
            target=ErrorTarget.CONVERSATION,
        )
    return [transform_fn(l) for l in lst if not is_none_or_nan(l)]


def is_none_or_nan(val: float) -> bool:
    """math.isnan raises an error if None is inputted. This is a more robust wrapper.

    :param val: The value to check.
    :type val: float
    :return: Whether the value is None or NaN.
    :rtype: bool
    """

    return val is None or math.isnan(val)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import os
from typing import Dict, Union, List

from typing_extensions import overload, override

from azure.ai.evaluation._model_configurations import Conversation
from azure.ai.evaluation._evaluators._common import PromptyEvaluatorBase


class RelevanceEvaluator(PromptyEvaluatorBase):
    """
    Evaluates relevance score for a given query and response or a multi-turn conversation, including reasoning.

    The relevance measure assesses the ability of answers to capture the key points of the context.
    High relevance scores signify the AI system's understanding of the input and its capability to produce coherent
    and contextually appropriate outputs. Conversely, low relevance scores indicate that generated responses might
    be off-topic, lacking in context, or insufficient in addressing the user's intended queries. Use the relevance
    metric when evaluating the AI system's performance in understanding the input and generating contextually
    appropriate responses.

    Relevance scores range from 1 to 5, with 1 being the worst and 5 being the best.

    :param model_config: Configuration for the Azure OpenAI model.
    :type model_config: Union[~azure.ai.evaluation.AzureOpenAIModelConfiguration,
        ~azure.ai.evaluation.OpenAIModelConfiguration]
    :param threshold: The threshold for the relevance evaluator. Default is 3.
    :type threshold: int

    .. admonition:: Example:

        .. literalinclude:: ../samples/evaluation_samples_evaluate.py
            :start-after: [START relevance_evaluator]
            :end-before: [END relevance_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call a RelevanceEvaluator with a query, response, and context.

    .. admonition:: Example using Azure AI Project URL:
        
        .. literalinclude:: ../samples/evaluation_samples_evaluate_fdp.py
            :start-after: [START relevance_evaluator]
            :end-before: [END relevance_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call RelevanceEvaluator using Azure AI Project URL in the following format 
                https://{resource_name}.services.ai.azure.com/api/projects/{project_name}

    .. admonition:: Example with Threshold:

        .. literalinclude:: ../samples/evaluation_samples_threshold.py
            :start-after: [START threshold_relevance_evaluator]
            :end-before: [END threshold_relevance_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize with threshold and call a RelevanceEvaluator with a query, response, and context.

    .. note::

        To align with our support of a diverse set of models, an output key without the `gpt_` prefix has been added.
        To maintain backwards compatibility, the old key with the `gpt_` prefix is still be present in the output;
        however, it is recommended to use the new key moving forward as the old key will be deprecated in the future.
    """

    # Constants must be defined within eval's directory to be save/loadable
    _PROMPTY_FILE = "relevance.prompty"
    _RESULT_KEY = "relevance"

    id = "azureml://registries/azureml/models/Relevance-Evaluator/versions/4"
    """Evaluator identifier, experimental and to be used only with evaluation in cloud."""

    @override
    def __init__(
        self, 
        model_config, 
        *,
        threshold=3
    ):
        current_dir = os.path.dirname(__file__)
        prompty_path = os.path.join(current_dir, self._PROMPTY_FILE)
        self._threshold = threshold
        self._higher_is_better = True
        super().__init__(
            model_config=model_config, 
            prompty_file=prompty_path, 
            result_key=self._RESULT_KEY, 
            threshold=threshold, 
            _higher_is_better=self._higher_is_better
        )

    @overload
    def __call__(
        self,
        *,
        query: str,
        response: str,
    ) -> Dict[str, Union[str, float]]:
        """Evaluate groundedness for given input of query, response, context

        :keyword query: The query to be evaluated.
        :paramtype query: str
        :keyword response: The response to be evaluated.
        :paramtype response: str
        :return: The relevance score.
        :rtype: Dict[str, float]
        """

    @overload
    def __call__(
        self,
        *,
        conversation: Conversation,
    ) -> Dict[str, Union[float, Dict[str, List[Union[str, float]]]]]:
        """Evaluate relevance for a conversation

        :keyword conversation: The conversation to evaluate. Expected to contain a list of conversation turns under the
            key "messages", and potentially a global context under the key "context". Conversation turns are expected
            to be dictionaries with keys "content", "role", and possibly "context".
        :paramtype conversation: Optional[~azure.ai.evaluation.Conversation]
        :return: The relevance score.
        :rtype: Dict[str, Union[float, Dict[str, List[float]]]]
        """

    @override
    def __call__(  # pylint: disable=docstring-missing-param
        self,
        *args,
        **kwargs,
    ):
        """Evaluate relevance. Accepts either a query and response for a single evaluation,
        or a conversation for a multi-turn evaluation. If the conversation has more than one turn,
        the evaluator will aggregate the results of each turn.

        :keyword query: The query to be evaluated. Mutually exclusive with the `conversation` parameter.
        :paramtype query: Optional[str]
        :keyword response: The response to be evaluated. Mutually exclusive with the `conversation` parameter.
        :paramtype response: Optional[str]
        :keyword conversation: The conversation to evaluate. Expected to contain a list of conversation turns under the
            key "messages", and potentially a global context under the key "context". Conversation turns are expected
            to be dictionaries with keys "content", "role", and possibly "context".
        :paramtype conversation: Optional[~azure.ai.evaluation.Conversation]
        :return: The relevance score.
        :rtype: Union[Dict[str, Union[str, float]], Dict[str, Union[float, Dict[str, List[Union[str, float]]]]]]
        """
        return super().__call__(*args, **kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import os
from typing import Dict, List, Optional, Union

from typing_extensions import overload, override
from azure.ai.evaluation._legacy._adapters._flows import AsyncPrompty

from azure.ai.evaluation._evaluators._common import PromptyEvaluatorBase
from azure.ai.evaluation._model_configurations import Conversation
from ..._common.utils import construct_prompty_model_config, validate_model_config

try:
    from ..._user_agent import USER_AGENT
except ImportError:
    USER_AGENT = "None"


class GroundednessEvaluator(PromptyEvaluatorBase[Union[str, float]]):
    """
    Evaluates groundedness score for a given query (optional), response, and context or a multi-turn conversation,
    including reasoning.

    The groundedness measure assesses the correspondence between claims in an AI-generated answer and the source
    context, making sure that these claims are substantiated by the context. Even if the responses from LLM are
    factually correct, they'll be considered ungrounded if they can't be verified against the provided sources
    (such as your input source or your database). Use the groundedness metric when you need to verify that
    AI-generated responses align with and are validated by the provided context.

    Groundedness scores range from 1 to 5, with 1 being the least grounded and 5 being the most grounded.

    :param model_config: Configuration for the Azure OpenAI model.
    :type model_config: Union[~azure.ai.evaluation.AzureOpenAIModelConfiguration,
        ~azure.ai.evaluation.OpenAIModelConfiguration]
    :param threshold: The threshold for the groundedness evaluator. Default is 3.
    :type threshold: int
        
    .. admonition:: Example:

        .. literalinclude:: ../samples/evaluation_samples_evaluate.py
            :start-after: [START groundedness_evaluator]
            :end-before: [END groundedness_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call a GroundednessEvaluator.

    .. admonition:: Example with Threshold:
        .. literalinclude:: ../samples/evaluation_samples_threshold.py
            :start-after: [START threshold_groundedness_evaluator]
            :end-before: [END threshold_groundedness_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize with threshold and call a GroundednessEvaluator.

    .. admonition:: Example using Azure AI Project URL:
        
        .. literalinclude:: ../samples/evaluation_samples_evaluate_fdp.py
            :start-after: [START groundedness_evaluator]
            :end-before: [END groundedness_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call GroundednessEvaluator using Azure AI Project URL in the following format 
                https://{resource_name}.services.ai.azure.com/api/projects/{project_name}

    .. note::

        To align with our support of a diverse set of models, an output key without the `gpt_` prefix has been added.
        To maintain backwards compatibility, the old key with the `gpt_` prefix is still be present in the output;
        however, it is recommended to use the new key moving forward as the old key will be deprecated in the future.
    """

    _PROMPTY_FILE_NO_QUERY = "groundedness_without_query.prompty"
    _PROMPTY_FILE_WITH_QUERY = "groundedness_with_query.prompty"
    _RESULT_KEY = "groundedness"
    _OPTIONAL_PARAMS = ["query"]

    id = "azureml://registries/azureml/models/Groundedness-Evaluator/versions/4"
    """Evaluator identifier, experimental and to be used only with evaluation in cloud."""

    @override
    def __init__(self, model_config, *, threshold=3, **kwargs):
        current_dir = os.path.dirname(__file__)
        prompty_path = os.path.join(current_dir, self._PROMPTY_FILE_NO_QUERY)  # Default to no query

        self._higher_is_better = True
        super().__init__(
            model_config=model_config,
            prompty_file=prompty_path,
            result_key=self._RESULT_KEY,
            threshold=threshold,
            _higher_is_better=self._higher_is_better
        )
        self._model_config = model_config
        self.threshold = threshold
        # Needs to be set because it's used in call method to re-validate prompt if `query` is provided

    @overload
    def __call__(
        self,
        *,
        response: str,
        context: str,
        query: Optional[str] = None,
    ) -> Dict[str, Union[str, float]]:
        """Evaluate groundedness for given input of response, context

        :keyword response: The response to be evaluated.
        :paramtype response: str
        :keyword context: The context to be evaluated.
        :paramtype context: str
        :keyword query: The query to be evaluated. Optional parameter for use with the `response`
            and `context` parameters. If provided, a different prompt template will be used for evaluation.
        :paramtype query: Optional[str]
        :return: The groundedness score.
        :rtype: Dict[str, float]
        """

    @overload
    def __call__(
        self,
        *,
        conversation: Conversation,
    ) -> Dict[str, Union[float, Dict[str, List[Union[str, float]]]]]:
        """Evaluate groundedness for a conversation

        :keyword conversation: The conversation to evaluate. Expected to contain a list of conversation turns under the
            key "messages", and potentially a global context under the key "context". Conversation turns are expected
            to be dictionaries with keys "content", "role", and possibly "context".
        :paramtype conversation: Optional[~azure.ai.evaluation.Conversation]
        :return: The groundedness score.
        :rtype: Dict[str, Union[float, Dict[str, List[float]]]]
        """

    @override
    def __call__(  # pylint: disable=docstring-missing-param
        self,
        *args,
        **kwargs,
    ):
        """Evaluate groundedness. Accepts either a query, response, and context for a single evaluation,
        or a conversation for a multi-turn evaluation. If the conversation has more than one turn,
        the evaluator will aggregate the results of each turn.

        :keyword query: The query to be evaluated. Mutually exclusive with `conversation`. Optional parameter for use
            with the `response` and `context` parameters. If provided, a different prompt template will be used for
            evaluation.
        :paramtype query: Optional[str]
        :keyword response: The response to be evaluated. Mutually exclusive with the `conversation` parameter.
        :paramtype response: Optional[str]
        :keyword context: The context to be evaluated. Mutually exclusive with the `conversation` parameter.
        :paramtype context: Optional[str]
        :keyword conversation: The conversation to evaluate. Expected to contain a list of conversation turns under the
            key "messages", and potentially a global context under the key "context". Conversation turns are expected
            to be dictionaries with keys "content", "role", and possibly "context".
        :paramtype conversation: Optional[~azure.ai.evaluation.Conversation]
        :return: The relevance score.
        :rtype: Union[Dict[str, Union[str, float]], Dict[str, Union[float, Dict[str, List[Union[str, float]]]]]]
        """

        if kwargs.get("query", None):
            current_dir = os.path.dirname(__file__)
            prompty_path = os.path.join(current_dir, self._PROMPTY_FILE_WITH_QUERY)
            self._prompty_file = prompty_path
            prompty_model_config = construct_prompty_model_config(
                validate_model_config(self._model_config),
                self._DEFAULT_OPEN_API_VERSION,
                USER_AGENT,
            )
            self._flow = AsyncPrompty.load(source=self._prompty_file, model=prompty_model_config)

        return super().__call__(*args, **kwargs)



from typing import Any, Mapping
import pytest
import logging
from azure.core.credentials import AzureSasCredential, TokenCredential, AccessToken
from azure.ai.evaluation._azure._clients import LiteMLClient


@pytest.mark.usefixtures("model_config", "project_scope", "recording_injection", "recorded_test")
class TestLiteAzureManagementClient(object):
    """End to end tests for the lite Azure management client."""

    @pytest.mark.azuretest
    def test_get_credential(self, project_scope, azure_cred):
        client = LiteMLClient(
            subscription_id=project_scope["subscription_id"],
            resource_group=project_scope["resource_group_name"],
            credential=azure_cred,
            logger=logging.getLogger(__name__),
        )

        credential = client.get_credential()
        assert isinstance(credential, TokenCredential)

    @pytest.mark.azuretest
    def test_get_token(self, project_scope, azure_cred):
        client = LiteMLClient(
            subscription_id=project_scope["subscription_id"],
            resource_group=project_scope["resource_group_name"],
            credential=azure_cred,
            logger=logging.getLogger(__name__),
        )

        token: AccessToken = client.get_token()
        assert isinstance(token, AccessToken) and len(token.token) > 0

    @pytest.mark.azuretest
    @pytest.mark.parametrize("include_credentials", [False, True])
    @pytest.mark.parametrize("config_name", ["sas", "none"])
    def test_workspace_get_default_store(
        self, azure_cred, datastore_project_scopes, config_name: str, include_credentials: bool
    ):
        project_scope = datastore_project_scopes[config_name]

        client = LiteMLClient(
            subscription_id=project_scope["subscription_id"],
            resource_group=project_scope["resource_group_name"],
            credential=azure_cred,
            logger=logging.getLogger(__name__),
        )

        store = client.workspace_get_default_datastore(
            workspace_name=project_scope["project_name"], include_credentials=include_credentials
        )

        assert store
        assert store.name
        assert store.account_name
        assert store.endpoint
        assert store.container_name
        if include_credentials:
            assert (
                (config_name == "account_key" and isinstance(store.credential, str))
                or (config_name == "sas" and isinstance(store.credential, AzureSasCredential))
                or (config_name == "none" and isinstance(store.credential, TokenCredential))
            )
        else:
            assert store.credential == None

    @pytest.mark.azuretest
    @pytest.mark.parametrize("config_name", ["sas", "none", "private"])
    def test_workspace_get_info(
        self, datastore_project_scopes: Mapping[str, Any], azure_cred: TokenCredential, config_name: str
    ):
        project_scope = datastore_project_scopes[config_name]

        client = LiteMLClient(
            subscription_id=project_scope["subscription_id"],
            resource_group=project_scope["resource_group_name"],
            credential=azure_cred,
            logger=logging.getLogger(__name__),
        )

        workspace = client.workspace_get_info(project_scope["project_name"])

        assert workspace
        assert workspace.name
        assert workspace.ml_flow_tracking_uri



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import os
import math
from typing import Dict, Union, List, Optional

from typing_extensions import overload, override

from azure.ai.evaluation._exceptions import EvaluationException, ErrorBlame, ErrorCategory, ErrorTarget
from azure.ai.evaluation._evaluators._common import PromptyEvaluatorBase
from azure.ai.evaluation._common.utils import parse_quality_evaluator_reason_score
from azure.ai.evaluation._model_configurations import Message
from azure.ai.evaluation._common._experimental import experimental

@experimental
class TaskAdherenceEvaluator(PromptyEvaluatorBase[Union[str, float]]):
    """The Task Adherence evaluator assesses how well an AI-generated response follows the assigned task based on:

        - Alignment with instructions and definitions
        - Accuracy and clarity of the response
        - Proper use of provided tool definitions

    Scoring is based on five levels:
    1. Fully Inadherent - Response completely ignores instructions.
    2. Barely Adherent - Partial alignment with critical gaps.
    3. Moderately Adherent - Meets core requirements but lacks precision.
    4. Mostly Adherent - Clear and accurate with minor issues.
    5. Fully Adherent - Flawless adherence to instructions.

    The evaluation includes a step-by-step reasoning process, a brief explanation, and a final integer score.


    :param model_config: Configuration for the Azure OpenAI model.
    :type model_config: Union[~azure.ai.evaluation.AzureOpenAIModelConfiguration,
        ~azure.ai.evaluation.OpenAIModelConfiguration]

    .. admonition:: Example:
        .. literalinclude:: ../samples/evaluation_samples_evaluate.py
            :start-after: [START task_adherence_evaluator]
            :end-before: [END task_adherence_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call an TaskAdherenceEvaluator with a query and response.
    
    .. admonition:: Example using Azure AI Project URL:
        
        .. literalinclude:: ../samples/evaluation_samples_evaluate_fdp.py
            :start-after: [START task_adherence_evaluator]
            :end-before: [END task_adherence_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call TaskAdherenceEvaluator using Azure AI Project URL in the following format 
                https://{resource_name}.services.ai.azure.com/api/projects/{project_name}

    """

    _PROMPTY_FILE = "task_adherence.prompty"
    _RESULT_KEY = "task_adherence"
    _OPTIONAL_PARAMS = ["tool_definitions"]

    _DEFAULT_TASK_ADHERENCE_SCORE = 3

    id = None
    """Evaluator identifier, experimental and to be used only with evaluation in cloud."""

    @override
    def __init__(self, model_config, *, threshold=_DEFAULT_TASK_ADHERENCE_SCORE,
                 **kwargs):
        current_dir = os.path.dirname(__file__)
        prompty_path = os.path.join(current_dir, self._PROMPTY_FILE)
        self.threshold = threshold
        super().__init__(model_config=model_config, prompty_file=prompty_path,
                         result_key=self._RESULT_KEY,
                         **kwargs)

    @overload
    def __call__(
        self,
        *,
        query: Union[str, List[dict]],
        response: Union[str, List[dict]],
        tool_definitions: Optional[Union[dict, List[dict]]] = None,
    ) -> Dict[str, Union[str, float]]:
        """Evaluate task adherence for a given query, response, and optional tool defintions.
        The query and response can be either a string or a list of messages.

        
        Example with string inputs and no tools:
            evaluator = TaskAdherenceEvaluator(model_config)
            query = "What is the weather today?"
            response = "The weather is sunny."

            result = evaluator(query=query, response=response)

        Example with list of messages:
            evaluator = TaskAdherenceEvaluator(model_config)
            query = [{'role': 'system', 'content': 'You are a friendly and helpful customer service agent.'}, {'createdAt': 1700000060, 'role': 'user', 'content': [{'type': 'text', 'text': 'Hi, I need help with the last 2 orders on my account #888. Could you please update me on their status?'}]}]
            response = [{'createdAt': 1700000070, 'run_id': '0', 'role': 'assistant', 'content': [{'type': 'text', 'text': 'Hello! Let me quickly look up your account details.'}]}, {'createdAt': 1700000075, 'run_id': '0', 'role': 'assistant', 'content': [{'type': 'tool_call', 'tool_call': {'id': 'tool_call_20250310_001', 'type': 'function', 'function': {'name': 'get_orders', 'arguments': {'account_number': '888'}}}}]}, {'createdAt': 1700000080, 'run_id': '0', 'tool_call_id': 'tool_call_20250310_001', 'role': 'tool', 'content': [{'type': 'tool_result', 'tool_result': '[{ "order_id": "123" }, { "order_id": "124" }]'}]}, {'createdAt': 1700000085, 'run_id': '0', 'role': 'assistant', 'content': [{'type': 'text', 'text': 'Thanks for your patience. I see two orders on your account. Let me fetch the details for both.'}]}, {'createdAt': 1700000090, 'run_id': '0', 'role': 'assistant', 'content': [{'type': 'tool_call', 'tool_call': {'id': 'tool_call_20250310_002', 'type': 'function', 'function': {'name': 'get_order', 'arguments': {'order_id': '123'}}}}, {'type': 'tool_call', 'tool_call': {'id': 'tool_call_20250310_003', 'type': 'function', 'function': {'name': 'get_order', 'arguments': {'order_id': '124'}}}}]}, {'createdAt': 1700000095, 'run_id': '0', 'tool_call_id': 'tool_call_20250310_002', 'role': 'tool', 'content': [{'type': 'tool_result', 'tool_result': '{ "order": { "id": "123", "status": "shipped", "delivery_date": "2025-03-15" } }'}]}, {'createdAt': 1700000100, 'run_id': '0', 'tool_call_id': 'tool_call_20250310_003', 'role': 'tool', 'content': [{'type': 'tool_result', 'tool_result': '{ "order": { "id": "124", "status": "delayed", "expected_delivery": "2025-03-20" } }'}]}, {'createdAt': 1700000105, 'run_id': '0', 'role': 'assistant', 'content': [{'type': 'text', 'text': 'The order with ID 123 has been shipped and is expected to be delivered on March 15, 2025. However, the order with ID 124 is delayed and should now arrive by March 20, 2025. Is there anything else I can help you with?'}]}]
            tool_definitions = [{'name': 'get_orders', 'description': 'Get the list of orders for a given account number.', 'parameters': {'type': 'object', 'properties': {'account_number': {'type': 'string', 'description': 'The account number to get the orders for.'}}}}, {'name': 'get_order', 'description': 'Get the details of a specific order.', 'parameters': {'type': 'object', 'properties': {'order_id': {'type': 'string', 'description': 'The order ID to get the details for.'}}}}, {'name': 'initiate_return', 'description': 'Initiate the return process for an order.', 'parameters': {'type': 'object', 'properties': {'order_id': {'type': 'string', 'description': 'The order ID for the return process.'}}}}, {'name': 'update_shipping_address', 'description': 'Update the shipping address for a given account.', 'parameters': {'type': 'object', 'properties': {'account_number': {'type': 'string', 'description': 'The account number to update.'}, 'new_address': {'type': 'string', 'description': 'The new shipping address.'}}}}]

            result = evaluator(query=query, response=response, tool_definitions=tool_definitions)

        :keyword query: The query being evaluated, either a string or a list of messages.
        :paramtype query: Union[str, List[dict]]
        :keyword response: The response being evaluated, either a string or a list of messages (full agent response potentially including tool calls)
        :paramtype response: Union[str, List[dict]]
        :keyword tool_definitions: An optional list of messages containing the tool definitions the agent is aware of.
        :paramtype tool_definitions: Optional[Union[dict, List[dict]]]
        :return: A dictionary with the task adherence evaluation results.
        :rtype: Dict[str, Union[str, float]]
        """

    @override
    def __call__(  # pylint: disable=docstring-missing-param
            self,
            *args,
            **kwargs,
    ):
        """
        Invokes the instance using the overloaded __call__ signature.

        For detailed parameter types and return value documentation, see the overloaded __call__ definition.
        """
        return super().__call__(*args, **kwargs)

    @override
    async def _do_eval(self, eval_input: Dict) -> Dict[str, Union[float, str]]:  # type: ignore[override]
        """Do Task Adherence evaluation.
        :param eval_input: The input to the evaluator. Expected to contain whatever inputs are needed for the _flow method
        :type eval_input: Dict
        :return: The evaluation result.
        :rtype: Dict
        """
        # we override the _do_eval method as we want the output to be a dictionary,
        # which is a different schema than _base_prompty_eval.py
        if "query" not in eval_input and "response" not in eval_input:
            raise EvaluationException(
                message=f"Both query and response must be provided as input to the Task Adherence evaluator.",
                internal_message=f"Both query and response must be provided as input to the Task Adherence evaluator.",
                blame=ErrorBlame.USER_ERROR,
                category=ErrorCategory.MISSING_FIELD,
                target=ErrorTarget.TASK_ADHERENCE_EVALUATOR,
            )

        llm_output = await self._flow(timeout=self._LLM_CALL_TIMEOUT, **eval_input)

        score = math.nan
        if llm_output:
            score, reason = parse_quality_evaluator_reason_score(llm_output, valid_score_range="[1-5]")

            score_result = 'pass' if score >= self.threshold else 'fail'

            return {
                f"{self._result_key}": score,
                f"{self._result_key}_result": score_result,
                f"{self._result_key}_threshold": self.threshold,
                f"{self._result_key}_reason": reason,
            }

        return {self._result_key: math.nan}




# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# NOTE: This is a direct port of the bare minimum needed for BatchEngine functionality from
#       the original Promptflow code. The goal here is expediency, not elegance. As such
#       parts of this code may be a little "quirky", seem incomplete in places, or contain
#       longer TODOs comments than usual. In a future code update, large swaths of this code
#       will be refactored or deleted outright.



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from ._document_retrieval import DocumentRetrievalEvaluator, RetrievalGroundTruthDocument, RetrievedDocument

__all__ = [
    "DocumentRetrievalEvaluator",
    "RetrievalGroundTruthDocument",
    "RetrievedDocument"
]
